from typing import Any, Optional, Tuple, List

import solders.system_program as sp
import solders.sysvar
from podite import pod, U64, U128, Option, I8, Static
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.rpc import types
from solana.rpc.commitment import Confirmed
from solana.transaction import Transaction
from solders.instruction import AccountMeta
from spl.token.constants import MINT_LEN
from spl.token.instructions import (
    TOKEN_PROGRAM_ID,
    create_associated_token_account,
    get_associated_token_address,
    mint_to_checked,
    MintToCheckedParams,
)

import dexteritysdk.codegen.dex.instructions as ixs
import dexteritysdk.codegen.dex.types as dex_types
import dexteritysdk.dex.addrs as dex_addrs
import dexteritysdk.program_ids as pids
from dexteritysdk.codegen.dex.types import MarketProductGroup, TraderRiskGroup
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.dex.addrs import crush
from dexteritysdk.program_ids import DEX_PROGRAM_ID
from dexteritysdk.utils import create_market_aob_ix
from dexteritysdk.utils.aob.state.base import SelfTradeBehavior, Side
from dexteritysdk.utils.aob.state.market_state import MarketState
from dexteritysdk.utils.aob.state.slab import Slab
from dexteritysdk.utils.solana import (
    actionify,
    Context,
    fetch_account_details,
    send_transaction,
    sighash_int,
)

DEFAULT_ORDERBOOK_SIZE = MarketState.calc_size()
DEFAULT_EVENT_QUEUE_SIZE = 100_000
DEFAULT_ASKS_SIZE = 100_000
DEFAULT_BIDS_SIZE = 100_000
DEFAULT_TICK_SIZE = 0.1
DEFAULT_DECIMALS = 6
DEFAULT_MATCH_LIMIT = 50
DEFAULT_OFFSET = 0
OUT_REGISTER_RISK_SIZE = 440  # fixed size taken from rust

PROGRAMS_SEED = f"{pids.AOB_PROGRAM_ID}:{pids.DEX_PROGRAM_ID}"
MARKET_PRODUCT_GROUP_SEED_LAYOUT = f"prod_grp:{PROGRAMS_SEED}:{{seed}}"
ORDERBOOK_SEED_LAYOUT = f"prod:ob:{PROGRAMS_SEED}:{{group}}:{{key}}"
EVENT_QUEUE_SEED_LAYOUT = f"prod:eq:{PROGRAMS_SEED}:{{group}}:{{key}}"
BIDS_SEED_LAYOUT = f"prod:bid:{PROGRAMS_SEED}:{{group}}:{{key}}"
ASKS_SEED_LAYOUT = f"prod:ask:{PROGRAMS_SEED}:{{group}}:{{key}}"
TRADER_RISK_GROUP_SEED_LAYOUT = f"trdr_grp:{PROGRAMS_SEED}:{{market_product_group}}"
MINT_SEED_LAYOUT = f"mint:{PROGRAMS_SEED}:{{seed}}"
OUT_REGISTER_RISK_LAYOUT = f"out_register_risk:{PROGRAMS_SEED}:{{group}}"
FEE_REGISTER_LAYOUT = f"fee:{PROGRAMS_SEED}:{{group}}"
RISK_CONFIG_LAYOUT = f"risk_config:{PROGRAMS_SEED}:{{group}}"
IN_REGISTER_RISK_LAYOUT = f"in_register_risk:{PROGRAMS_SEED}:{{group}}"


def _to_rust_string_for_construct(s: str):
    return {"length": len(s), "chars": s}


def _calc_rent(space, client=None):
    if client is None:
        client = Context.get_global_client()
    return client.get_minimum_balance_for_rent_exemption(space)["result"]


def _post_create_market_product_group(resp):
    addr = resp.instructions[0]["accounts"][1]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify(post_process=_post_create_market_product_group)
def create_market_product_group(
        authority: Pubkey,
        seed: str,
        vault_mint: Pubkey,
        fee_collector: Pubkey,
        fee_model_configuration_acct: Pubkey,
        name: Optional[str] = None,
        risk_engine_program: Optional[Pubkey] = pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
        fee_model_program: Optional[Pubkey] = pids.CONSTANT_FEES_MODEL_PROGRAM_ID,
):
    market_product_group = dex_addrs.get_market_product_group_addr(authority, seed)
    space = Static[MarketProductGroup].calc_size() + 8
    rent = _calc_rent(space)

    return Transaction(fee_payer=authority).add(
        sp.create_account_with_seed(
            sp.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=market_product_group,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(crush(MARKET_PRODUCT_GROUP_SEED_LAYOUT.format(seed=seed))),
                lamports=rent,
                space=space,
                program_id=pids.DEX_PROGRAM_ID,
            )
        )
    )


def _post_init_market_product_group(resp):
    addr = resp.instructions[0]["accounts"][1]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify(post_process=_post_init_market_product_group)
def init_market_product_group(
        authority: Pubkey,
        seed: str,
        vault_mint: Pubkey,
        vault: Pubkey,
        fee_collector: Pubkey,
        fee_model_configuration_acct: Pubkey,
        risk_model_configuration_acct: Pubkey,
        name: Optional[str] = None,
        risk_engine_program: Optional[Pubkey] = pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
        fee_model_program: Optional[Pubkey] = pids.CONSTANT_FEES_MODEL_PROGRAM_ID,
        is_risk_anchor: bool = True,
        params: dex_types.InitializeMarketProductGroupParams = None
):
    market_product_group = dex_addrs.get_market_product_group_addr(authority, seed)

    fee_output_register = dex_addrs.get_fee_register_addr(
        authority, market_product_group, fee_model_program, FEE_REGISTER_LAYOUT
    )
    risk_output_register = dex_addrs.get_risk_register_addr(
        authority, market_product_group, risk_engine_program, OUT_REGISTER_RISK_LAYOUT
    )

    if name is None:
        name = seed
    name = bytes(name, encoding="utf-8")
    if len(name) > 16:
        name = name[:16]
    elif len(name) < 16:
        name = name + b" " * (16 - len(name))

    if params is None:
        if is_risk_anchor:
            disc_len = U64(8)
            health = U64(sighash_int("validate_account_health"))
            liquidation = U64(sighash_int("validate_account_liquidation"))
            create_risk_state = U64(sighash_int("create_risk_state_account"))
        else:
            disc_len = U64(1)
            health = U64(0)
            liquidation = U64(1)
            create_risk_state = U64(2)
        params = dex_types.InitializeMarketProductGroupParams(
            name=name,
            validate_account_discriminant_len=disc_len,
            find_fees_discriminant_len=1,
            validate_account_health_discriminant=U64.to_bytes(health),
            validate_account_liquidation_discriminant=U64.to_bytes(liquidation),
            create_risk_state_account_discriminant=U64.to_bytes(create_risk_state),
            find_fees_discriminant=U64.to_bytes(0),
            max_maker_fee_bps=1000,
            min_maker_fee_bps=-10,
            max_taker_fee_bps=1000,
            min_taker_fee_bps=0,
        )

    return Transaction(fee_payer=authority).add(
        ixs.initialize_market_product_group(
            authority=authority,
            market_product_group=market_product_group,
            vault_mint=vault_mint,
            vault=vault,
            fee_collector=fee_collector,
            fee_model_program=fee_model_program,
            fee_model_configuration_acct=fee_model_configuration_acct,
            risk_engine_program=risk_engine_program,
            fee_output_register=fee_output_register,
            risk_model_configuration_acct=risk_model_configuration_acct,
            risk_output_register=risk_output_register,
            sysvar_rent=solders.sysvar.SYSVAR_RENT_PUBKEY,
            params=params,
        ),
    )


def _post_create_risk_register(resp):
    addr = resp.instructions[0]["accounts"][1]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


def _post_create_risk_config(resp):
    addr = resp.instructions[0]["accounts"][1]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify
def create_fee_register(
        authority: Pubkey,
        group: Pubkey,
        register_size: Any,
        program_id: Optional[Pubkey] = pids.CONSTANT_FEES_MODEL_PROGRAM_ID,
        layout_str: Optional[str] = FEE_REGISTER_LAYOUT,
):
    register_info = dex_addrs.get_fee_register_addr(
        authority, group, program_id, layout_str
    )

    return Transaction(fee_payer=authority).add(
        system_program.create_account_with_seed(
            system_program.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=register_info,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(
                    dex_addrs.crush(layout_str.format(group=group))
                ),
                space=register_size,
                lamports=_calc_rent(register_size),
                program_id=program_id,
            )
        ),
    )


from solders import system_program


@actionify(post_process=_post_create_risk_config)
def create_risk_config_acct(
        authority: Pubkey,
        group: Pubkey,
        program_id: Optional[Pubkey] = pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
        layout_str: Optional[str] = RISK_CONFIG_LAYOUT,
):
    risk_config_acct = dex_addrs.get_risk_register_addr(
        authority, group, program_id, layout_str
    )

    if program_id == pids.ALPHA_RISK_ENGINE_PROGRAM_ID:
        register_size = 0
    else:
        print("Only alpha risk engine initialization currently implemented")
        raise Exception("Only alpha risk engine initialization currently implemented")

    return Transaction(fee_payer=authority).add(
        system_program.create_account_with_seed(
            system_program.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=risk_config_acct,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(
                    dex_addrs.crush(layout_str.format(group=group))
                ),
                space=register_size,
                lamports=_calc_rent(register_size),
                program_id=program_id,
            )
        )
    )


@actionify(post_process=_post_create_risk_register)
def create_risk_register(
        authority: Pubkey,
        group: Pubkey,
        register_size: Any,
        program_id: Optional[Pubkey] = pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
        layout_str: Optional[str] = OUT_REGISTER_RISK_LAYOUT,
):
    register_info = dex_addrs.get_risk_register_addr(
        authority, group, program_id, layout_str
    )

    return Transaction(fee_payer=authority).add(
        system_program.create_account_with_seed(
            system_program.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=register_info,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(
                    dex_addrs.crush(layout_str.format(group=group))
                ),
                space=register_size,
                lamports=_calc_rent(register_size),
                program_id=program_id,
            ),
        ),
    )


def create_risk_model_configuration_acct(
        authority: Keypair,
        program_id: Optional[Pubkey] = pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
):
    risk_model_key = Keypair()
    if program_id == pids.NOOP_RISK_ENGINE_PROGRAM_ID:
        space = 0
    elif program_id == pids.ALPHA_RISK_ENGINE_PROGRAM_ID:
        space = 0
    else:
        print("Unexpected program ID")
        raise Exception

    rent = _calc_rent(space)
    txn = Transaction().add(
        sp.create_account(
            sp.CreateAccountParams(
                from_pubkey=authority.pubkey(),
                new_account_pubkey=risk_model_key.pubkey(),
                lamports=rent,
                space=space,
                program_id=program_id,
            )
        )
    )

    send_transaction(
        txn,
        authority,
        risk_model_key,
        opts=types.TxOpts(
            skip_preflight=False,
            skip_confirmation=False,
            preflight_commitment=Confirmed,
        ),
    )
    return risk_model_key.pubkey


def get_orderbook_addrs(
        authority: Pubkey,
        group: Pubkey,
        key: Pubkey,
):
    return (
        Pubkey.create_with_seed(
            base=authority,
            seed=crush(ORDERBOOK_SEED_LAYOUT.format(group=group, key=key)),
            program_id=pids.AOB_PROGRAM_ID,
        ),
        Pubkey.create_with_seed(
            base=authority,
            seed=crush(EVENT_QUEUE_SEED_LAYOUT.format(group=group, key=key)),
            program_id=pids.AOB_PROGRAM_ID,
        ),
        Pubkey.create_with_seed(
            base=authority,
            seed=crush(BIDS_SEED_LAYOUT.format(group=group, key=key)),
            program_id=pids.AOB_PROGRAM_ID,
        ),
        Pubkey.create_with_seed(
            base=authority,
            seed=crush(ASKS_SEED_LAYOUT.format(group=group, key=key)),
            program_id=pids.AOB_PROGRAM_ID,
        ),
    )


def create_aob_orderbook_helper(
        authority: Pubkey,
        market_product_group: Pubkey,
        product_key: Pubkey,
        orderbook_authority: Pubkey,
        callback_info_len: int = 32,
        callback_id_len: int = 32,
        min_base_order_size: int = 1,
        orderbook_size: int = DEFAULT_ORDERBOOK_SIZE,
        event_queue_size: int = DEFAULT_EVENT_QUEUE_SIZE,
        asks_size: int = DEFAULT_ASKS_SIZE,
        bids_size: int = DEFAULT_BIDS_SIZE,
):
    orderbook, event_queue, bids, asks = get_orderbook_addrs(
        authority, market_product_group, product_key
    )
    return Transaction(fee_payer=authority).add(
        sp.create_account_with_seed(
            sp.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=orderbook,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(crush(
                    ORDERBOOK_SEED_LAYOUT.format(
                        group=market_product_group, key=product_key
                    )
                )),
                lamports=_calc_rent(orderbook_size),
                space=orderbook_size,
                program_id=pids.AOB_PROGRAM_ID,
            )
        ),
        sp.create_account_with_seed(
            sp.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=event_queue,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(crush(
                    EVENT_QUEUE_SEED_LAYOUT.format(
                        group=market_product_group, key=product_key
                    )
                )),
                lamports=_calc_rent(event_queue_size),
                space=event_queue_size,
                program_id=pids.AOB_PROGRAM_ID,
            )),
        sp.create_account_with_seed(
            sp.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=bids,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(crush(
                    BIDS_SEED_LAYOUT.format(group=market_product_group, key=product_key)
                )),
                lamports=_calc_rent(bids_size),
                space=bids_size,
                program_id=pids.AOB_PROGRAM_ID,
            )),
        sp.create_account_with_seed(
            sp.CreateAccountWithSeedParams(
                from_pubkey=authority,
                new_account_pubkey=asks,
                base_pubkey=authority,
                seed=_to_rust_string_for_construct(crush(
                    ASKS_SEED_LAYOUT.format(group=market_product_group, key=product_key)
                )),
                lamports=_calc_rent(asks_size),
                space=asks_size,
                program_id=pids.AOB_PROGRAM_ID,
            )),
        create_market_aob_ix(
            market=orderbook,
            event_queue=event_queue,
            bids=bids,
            asks=asks,
            caller_authority=orderbook_authority,
            callback_info_len=callback_info_len,
            callback_id_len=callback_id_len,
            min_base_order_size=min_base_order_size,
        ),
    )


def get_agnostic_orderbook_authority(product_key: Pubkey):
    return Pubkey.find_program_address([bytes(product_key)], DEX_PROGRAM_ID)[0]


@actionify
def create_market_product(
        authority: Pubkey,
        market_product_group: Pubkey,
        product_key: Pubkey,
        name: str,
        callback_info_len: int = 32,
        callback_id_len: int = 32,
        min_base_order_size: int = 1,
        orderbook_size: int = DEFAULT_ORDERBOOK_SIZE,
        event_queue_size: int = DEFAULT_EVENT_QUEUE_SIZE,
        asks_size: int = DEFAULT_ASKS_SIZE,
        bids_size: int = DEFAULT_BIDS_SIZE,
        tick_size: float = DEFAULT_TICK_SIZE,
        base_decimals: int = DEFAULT_DECIMALS,
        price_offset: float = DEFAULT_OFFSET,
):
    tick_size = Fractional.to_decimal(tick_size)
    price_offset = Fractional.to_decimal(price_offset)
    orderbook, _, _, _ = get_orderbook_addrs(
        authority, market_product_group, product_key
    )
    orderbook_authority = get_agnostic_orderbook_authority(product_key)
    return Transaction(fee_payer=authority).add(
        create_aob_orderbook_helper(
            authority,
            market_product_group,
            product_key,
            orderbook_authority,
            callback_info_len=callback_info_len,
            callback_id_len=callback_id_len,
            min_base_order_size=min_base_order_size,
            orderbook_size=orderbook_size,
            event_queue_size=event_queue_size,
            asks_size=asks_size,
            bids_size=bids_size,
        ),
        ixs.initialize_market_product(
            authority=authority,
            market_product_group=market_product_group,
            product=product_key,
            orderbook=orderbook,
            params=dex_types.InitializeMarketProductParams(
                name=name.encode("utf-8").ljust(16, b"\x00"),
                tick_size=tick_size,
                base_decimals=base_decimals,
                price_offset=price_offset,
            ),
        ),
    )


# Format of the seeds is [product_key_1, ..., product_key_N, [ratio_1, ..., ratio_N]]
def get_combo_product_key(product_keys, ratios):
    sort_idx = [i[0] for i in sorted(enumerate(product_keys), key=lambda x: str(x[1]))]
    product_keys = [product_keys[idx] for idx in sort_idx]
    ratios = [ratios[idx] for idx in sort_idx]
    seeds = [bytes(product_key) for product_key in product_keys] + [
        I8.to_bytes(r) for r in ratios
    ]
    return Pubkey.find_program_address(seeds, DEX_PROGRAM_ID)[0]


@actionify
def create_combo(
        authority: Pubkey,
        market_product_group: Pubkey,
        products: List[Pubkey],
        ratios: List[int],
        name: str,
        callback_info_len: int = 32,
        callback_id_len: int = 32,
        min_base_order_size: int = 1,
        orderbook_size: int = DEFAULT_ORDERBOOK_SIZE,
        event_queue_size: int = DEFAULT_EVENT_QUEUE_SIZE,
        asks_size: int = DEFAULT_ASKS_SIZE,
        bids_size: int = DEFAULT_BIDS_SIZE,
        tick_size: float = DEFAULT_TICK_SIZE,
        base_decimals: int = DEFAULT_DECIMALS,
        price_offset: float = DEFAULT_OFFSET,
):
    product_key = get_combo_product_key(products, ratios)
    orderbook_authority = get_agnostic_orderbook_authority(product_key)

    tick_size = Fractional.to_decimal(tick_size)
    price_offset = Fractional.to_decimal(price_offset)
    orderbook, _, _, _ = get_orderbook_addrs(
        authority, market_product_group, product_key
    )
    return Transaction(fee_payer=authority).add(
        create_aob_orderbook_helper(
            authority,
            market_product_group,
            product_key,
            orderbook_authority,
            callback_info_len=callback_info_len,
            callback_id_len=callback_id_len,
            min_base_order_size=min_base_order_size,
            orderbook_size=orderbook_size,
            event_queue_size=event_queue_size,
            asks_size=asks_size,
            bids_size=bids_size,
        ),
        ixs.initialize_combo(
            authority=authority,
            market_product_group=market_product_group,
            orderbook=orderbook,
            params=dex_types.InitializeComboParams(
                name=name.encode("utf-8").ljust(16, b"\x00"),
                ratios=ratios,
                tick_size=tick_size,
                base_decimals=base_decimals,
                price_offset=price_offset,
            ),
            remaining_accounts=[AccountMeta(pubkey=p, is_signer=False, is_writable=False) for p in products]
        ),
    )


def _post_create_trader_risk_group(resp):
    addr = resp.instructions[0]["accounts"][1]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


_risk_state_account_num = 0
@actionify(post_process=_post_create_trader_risk_group)
def create_trader_risk_group(
        trader: Pubkey,
        market_product_group: Pubkey,
        fee_ix: Optional[Any] = None,
        program_id: Optional[Pubkey] = pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
):
    trader_risk_group = dex_addrs.get_trader_risk_group_addr(trader, market_product_group)
    space = TraderRiskGroup.calc_size() + 8
    rent = _calc_rent(space)

    trader_fee_acct = dex_addrs.get_trader_fee_state_acct(trader_risk_group, market_product_group)

    if program_id == pids.ALPHA_RISK_ENGINE_PROGRAM_ID:
        register_size = alpha_state.Health.calc_max_size() + 8
    else:
        raise Exception("Not implemented")

    risk_state_account = Keypair()
    global _risk_state_account_num
    Context.add_signers((risk_state_account, "risk_state_account" + str(_risk_state_account_num)))
    _risk_state_account_num = _risk_state_account_num + 1

    if fee_ix is None:
        import dexteritysdk.constant_fees.instructions as fee_ixs

        fee_model_config_acct = dex_addrs.get_fee_model_configuration_addr(
            market_product_group, pids.CONSTANT_FEES_MODEL_PROGRAM_ID
        )
        fee_ix = fee_ixs.initialize_trader_acct_ix(
            pids.CONSTANT_FEES_MODEL_PROGRAM_ID,
            trader,
            fee_model_config_acct,
            trader_fee_acct,
            market_product_group,
            trader_risk_group=trader_risk_group,
            system_program=pids.SYSTEM_PROGRAM_ID,
        )

    return Transaction(fee_payer=trader).add(
        sp.create_account_with_seed(
            sp.CreateAccountWithSeedParams(
                from_pubkey=trader,
                new_account_pubkey=trader_risk_group,
                base_pubkey=trader,
                seed=_to_rust_string_for_construct(crush(
                    TRADER_RISK_GROUP_SEED_LAYOUT.format(
                        market_product_group=market_product_group
                    )
                )),
                lamports=rent,
                space=space,
                program_id=pids.DEX_PROGRAM_ID,
            )),
        fee_ix,
        ixs.initialize_trader_risk_group(
            owner=trader,
            trader_risk_group=trader_risk_group,
            market_product_group=market_product_group,
            risk_signer=dex_addrs.get_risk_signer(mpg=market_product_group),
            risk_state_account=risk_state_account.pubkey(),
            fee_state_account=trader_fee_acct,
            risk_engine_program=pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
        ),
    )


def _post_init_mint(resp):
    addr = resp.instructions[0]["accounts"][1]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        print(error_info)
        if error_ix == 0 and type(error_info) != str and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify(post_process=_post_init_mint)
def init_mint(
        authority: Pubkey,
        mint: Pubkey,
        mint_decimals: int = 6,
):
    from spl import token
    from spl.token import instructions as spl_ixs
    create_mint_ix = system_program.create_account(system_program.CreateAccountParams(
        from_pubkey=authority,
        new_account_pubkey=mint,
        lamports=1000_000_000,
        space=token.constants.MINT_LEN,
        program_id=token.constants.TOKEN_PROGRAM_ID,
    ))

    init_mint_ix = spl_ixs.initialize_mint(spl_ixs.InitializeMintParams(
        decimals=mint_decimals,
        freeze_authority=authority,
        mint=mint,
        mint_authority=authority,
        program_id=token.constants.TOKEN_PROGRAM_ID,
    ))
    return Transaction().add(create_mint_ix, init_mint_ix)


def _post_init_trader_mint_account(resp):
    addr = resp.instructions[0]["accounts"][1]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify(post_process=_post_init_trader_mint_account)
def init_trader_mint_account(
        trader: Pubkey,
        mint: Pubkey,
):
    return Transaction(fee_payer=trader).add(
        create_associated_token_account(trader, trader, mint)
    )


@actionify
def mint_to_trader(
        trader: Pubkey, mint: Pubkey, mint_authority: Pubkey, amount, decimals
):
    token_address = get_associated_token_address(trader, mint)
    return Transaction(fee_payer=trader).add(
        mint_to_checked(
            MintToCheckedParams(
                TOKEN_PROGRAM_ID,
                mint,
                token_address,
                mint_authority,
                amount,
                decimals,
            )
        )
    )


@pod
class DexOrderSummary:
    posted_order_id: Option[U128]
    total_base_qty: U64
    total_quote_qty: U64
    total_base_qty_posted: U64

    @classmethod
    def from_bytes_partial(cls, buffer: bytes) -> Tuple[object, bytes]:
        obj, remaining = super().from_bytes_partial(buffer)

        # TODO do any modification needed here (re fixed-point stuff)

        return obj, remaining


def _post_new_order(resp):
    if resp.error:
        return None, resp

    raw_summary = resp.emitted_logs["new-order:order-summary"]

    return DexOrderSummary.from_bytes(raw_summary), resp


@actionify(post_process=_post_new_order)
def new_order(
        trader: Pubkey,
        market_product_group: Pubkey,
        product_key: Pubkey,
        side: Side,
        limit_price: U64,
        max_base_qty: U64,
        order_type: dex_types.OrderType,
        trader_fee_acct: Optional[Pubkey] = None,
        fee_model_program: Optional[Pubkey] = pids.CONSTANT_FEES_MODEL_PROGRAM_ID,
        fee_register_acct: Optional[Pubkey] = None,
        risk_model_configuration_acct: Optional[Pubkey] = None,
        risk_register_acct: Optional[Pubkey] = None,
        risk_state_account: Optional[Pubkey] = None,
        self_trade_behavior: SelfTradeBehavior = SelfTradeBehavior.CANCEL_PROVIDE,
        match_limit: U64 = DEFAULT_MATCH_LIMIT,
        tick_size: Optional[U64] = None,
        decimals: Optional[U64] = None,
        authority: Optional[Pubkey] = None,
        risk_engine_program: Optional[Pubkey] = None,
):
    if (
            authority is None
            or decimals is None
            or tick_size is None
            or risk_engine_program is None
            or fee_model_program is None
    ):
        group_details = fetch_account_details(market_product_group)
        group_obj: MarketProductGroup = group_details.data_obj
        authority = group_obj.authority
        decimals = group_obj.decimals
        product = group_obj.get_product_by_key(
            product_key
        ) or group_obj.get_combo_by_key(product_key)
        tick_size = product.tick_size
        risk_engine_program = group_obj.risk_engine_program_id
        fee_model_program = group_obj.fee_model_program_id
        risk_model_configuration_acct = group_obj.risk_model_configuration_acct

    fee_model_configuration_acct = ixs.common.get_fee_model_configuration_addr(
        market_product_group,
        fee_model_program,
    )

    trader_risk_group = dex_addrs.get_trader_risk_group_addr(trader, market_product_group)
    if trader_fee_acct is None:
        trader_fee_acct = dex_addrs.get_trader_fee_state_acct(
            trader_risk_group,
            market_product_group,
            fee_model_program,
        )

    if risk_register_acct is None:
        out_register_acct = dex_addrs.get_risk_register_addr(
            authority,
            market_product_group,
            risk_engine_program,
            OUT_REGISTER_RISK_LAYOUT,
        )
    if risk_state_account is None:
        risk_state_account = dex_addrs.get_risk_register_addr(
            trader,
            trader_risk_group,
            risk_engine_program,
            IN_REGISTER_RISK_LAYOUT,
        )

    if fee_register_acct is None:
        fee_register_acct = dex_addrs.get_fee_register_addr(
            authority,
            market_product_group,
            fee_model_program,
            FEE_REGISTER_LAYOUT,
        )

    (orderbook, event_queue, bids, asks) = get_orderbook_addrs(
        authority,
        market_product_group,
        product_key,
    )
    limit_price = Fractional.to_decimal(limit_price)
    max_base_qty = Fractional.to_decimal(max_base_qty)

    return Transaction(fee_payer=trader).add(
        ixs.new_order(
            user=trader,
            trader_risk_group=trader_risk_group,
            market_product_group=market_product_group,
            product=product_key,
            orderbook=orderbook,
            event_queue=event_queue,
            bids=bids,
            asks=asks,
            params=dex_types.NewOrderParams(
                side=side,
                limit_price=limit_price,
                max_base_qty=max_base_qty,
                order_type=order_type,
                self_trade_behavior=self_trade_behavior,
                match_limit=match_limit,
            ),
            risk_engine_program=risk_engine_program,
            risk_model_configuration_acct=risk_model_configuration_acct,
            fee_model_configuration_acct=fee_model_configuration_acct,
            fee_output_register=fee_register_acct,
            trader_fee_state=trader_fee_acct,
            fee_model_program=fee_model_program,
            risk_output_register=out_register_acct,
            risk_state_account_info=risk_state_account,
        )
    )


@actionify
def cancel_order(
        trader: Pubkey,
        market_product_group: Pubkey,
        product_key: Pubkey,
        order_id: U128,
        authority: Optional[Pubkey] = None,
        risk_engine_program: Optional[Pubkey] = None,
        risk_register_acct: Optional[Pubkey] = None,
        risk_state_account: Optional[Pubkey] = None,
):
    if authority is None or risk_engine_program is None:
        group_details = fetch_account_details(market_product_group)
        group_obj: MarketProductGroup = group_details.data_obj
        authority = group_obj.authority
        risk_engine_program = group_obj.risk_engine_program_id

    trader_risk_group = dex_addrs.get_trader_risk_group_addr(trader, market_product_group)

    if risk_register_acct is None:
        risk_register_acct = dex_addrs.get_risk_register_addr(
            authority,
            market_product_group,
            risk_engine_program,
            OUT_REGISTER_RISK_LAYOUT,
        )

    if risk_state_account is None:
        risk_state_account = dex_addrs.get_risk_register_addr(
            trader,
            trader_risk_group,
            risk_engine_program,
            IN_REGISTER_RISK_LAYOUT,
        )

    (orderbook, event_queue, bids, asks) = get_orderbook_addrs(
        authority,
        market_product_group,
        product_key,
    )

    return Transaction(fee_payer=trader).add(
        ixs.cancel_order(
            user=trader,
            trader_risk_group=trader_risk_group,
            market_product_group=market_product_group,
            product=product_key,
            orderbook=orderbook,
            event_queue=event_queue,
            bids=bids,
            asks=asks,
            risk_output_register=risk_register_acct,
            risk_engine_program=risk_engine_program,
            risk_state_account_info=risk_state_account,
            aaob_program=pids.AOB_PROGRAM_ID,
            market_signer=authority,
            risk_model_configuration_acct=dex_addrs.get_risk_model_configuration_addr(market_product_group,
                                                                                      risk_engine_program),
            params=dex_types.CancelOrderParams(
                order_id=order_id,
            ),
        )
    )


@actionify
def consume_orderbook_events(
        market_product_group: Pubkey,
        product_key: Pubkey,
        reward_target: Pubkey,
        max_iterations: U64,
        risk_model_configuration_acct: Optional[Pubkey] = None,
        fee_register_acct: Optional[Pubkey] = None,
        user_accounts: Optional[List[Pubkey]] = None,
        authority: Optional[Pubkey] = None,
        fee_program: Pubkey = None,
        risk_output_register: Optional[Pubkey] = None,
        risk_engine_program: Optional[Pubkey] = None,
):
    if authority is None or fee_program is None or risk_engine_program is None:
        group_details = fetch_account_details(market_product_group)
        group_obj: MarketProductGroup = group_details.data_obj
        authority = group_obj.authority
        fee_program = group_obj.fee_model_program_id
        risk_engine_program = group_obj.risk_engine_program_id
        risk_model_configuration_acct = group_obj.risk_model_configuration_acct
        risk_output_register = group_obj.risk_output_register

    orderbook, event_queue, *_ = get_orderbook_addrs(
        authority,
        market_product_group,
        product_key,
    )

    if user_accounts is None:
        event_queue_details = fetch_account_details(event_queue)
        event_queue_obj: Slab = event_queue_details.data_obj

        if event_queue_obj.header.count == 0:
            raise RuntimeError("No event is in event_queue")

        trader_risk_groups = []
        n_events_to_process = min(max_iterations, event_queue_obj.header.count)
        for i in range(n_events_to_process):
            trader_risk_groups.extend(event_queue_obj[i].get_user_accounts())

        trader_risk_groups = list(
            map(Pubkey, set(list(map(bytes, trader_risk_groups))))
        )
        user_accounts = []
        for trader in trader_risk_groups:
            trader_fee_account = dex_addrs.get_trader_fee_state_acct(trader, market_product_group)
            trader_group_details = fetch_account_details(trader)
            trader_group_obj: TraderRiskGroup = trader_group_details.data_obj
            user_accounts.extend(
                [trader, trader_fee_account, trader_group_obj.risk_register]
            )

    if fee_register_acct is None:
        fee_register_acct = dex_addrs.get_risk_register_addr(
            authority,
            market_product_group,
            fee_program,
            FEE_REGISTER_LAYOUT,
        )

    risk_and_fee_signer, _ = Pubkey.find_program_address(
        seeds=[
            bytes(market_product_group),
        ],
        program_id=pids.DEX_PROGRAM_ID,
    )

    return Transaction(fee_payer=reward_target).add(
        ixs.consume_orderbook_events(
            market_product_group=market_product_group,
            product=product_key,
            orderbook=orderbook,
            event_queue=event_queue,
            reward_target=reward_target,
            fee_output_register=fee_register_acct,
            risk_and_fee_signer=risk_and_fee_signer,
            fee_model_program=fee_program,
            aaob_program=pids.AOB_PROGRAM_ID,
            fee_model_configuration_acct=dex_addrs.get_fee_model_configuration_addr(market_product_group),
            market_signer=authority,
            params=dex_types.ConsumeOrderbookEventsParams(
                max_iterations=max_iterations,
            ),
            remaining_accounts=user_accounts
        )
    )


@actionify
def deposit_funds(
        trader: Pubkey,
        trader_wallet: Pubkey,
        market_product_group: Pubkey,
        quantity: U64,
):
    trader_risk_group = dex_addrs.get_trader_risk_group_addr(trader, market_product_group)
    return Transaction(fee_payer=trader).add(
        ixs.deposit_funds(
            user=trader,
            user_token_account=trader_wallet,
            trader_risk_group=trader_risk_group,
            market_product_group=market_product_group,
            market_product_group_vault=dex_addrs.get_market_product_group_vault(market_product_group),
            params=dex_types.DepositFundsParams(
                quantity=Fractional.to_decimal(quantity),
            ),
        )
    )


@actionify
def withdraw_funds(
        trader: Pubkey,
        trader_wallet: Pubkey,
        market_product_group: Pubkey,
        quantity: U64,
):
    # todo fixme
    trader_risk_group = dex_addrs.get_trader_risk_group_addr(trader, market_product_group)
    return Transaction(fee_payer=trader).add(
        ixs.withdraw_funds(
            user=trader,
            user_token_account=trader_wallet,
            trader_risk_group=trader_risk_group,
            market_product_group=market_product_group,
            market_product_group_vault=dex_addrs.get_market_product_group_vault(mpg_key=market_product_group),
            params=dex_types.WithdrawFundsParams(
                quantity=Fractional.to_decimal(quantity),
            ),
        )
    )


@actionify
def update_trader_funding(
        market_product_group: Pubkey,
        trader: Optional[Pubkey] = None,
        trader_risk_group: Optional[Pubkey] = None,
        fee_payer: Optional[Pubkey] = None,  # note: anyone can call this
):
    # todo fixme
    if (trader is None and trader_risk_group is None) or (
            trader is not None and trader_risk_group is not None
    ):
        raise RuntimeError("Exactly one of trader and trader_risk_group must be passed")

    if trader_risk_group is None:
        trader_risk_group = dex_addrs.get_trader_risk_group_addr(trader, market_product_group)
    return Transaction(fee_payer=trader).add(
        ixs.update_trader_funding_ix(
            market_product_group=market_product_group,
            trader_risk_group=trader_risk_group,
        )
    )


@actionify
def clear_expired_orders(
        market_product_group: Pubkey,
        product_key: Pubkey,
        authority: Pubkey = None,
        num_orders_to_cancel: int = 5,
):
    # todo fixme
    if authority is None:
        group_details = fetch_account_details(market_product_group)
        group_obj: MarketProductGroup = group_details.data_obj
        authority = group_obj.authority
    (orderbook, event_queue, bids, asks) = get_orderbook_addrs(
        authority,
        market_product_group,
        product_key,
    )
    return Transaction(fee_payer=authority).add(
        ixs.clear_expired_orderbook_ix(
            market_product_group=market_product_group,
            product_key=product_key,
            orderbook=orderbook,
            event_queue=event_queue,
            bids=bids,
            asks=asks,
            num_orders_to_cancel=num_orders_to_cancel,
        )
    )


@actionify
def remove_market_product(
        market_product_group: Pubkey,
        product_key: Pubkey,
        authority: Pubkey = None,
):
    # todo fixme
    if authority is None:
        group_details = fetch_account_details(market_product_group)
        group_obj: MarketProductGroup = group_details.data_obj
        authority = group_obj.authority
    (orderbook, event_queue, bids, asks) = get_orderbook_addrs(
        authority,
        market_product_group,
        product_key,
    )
    return Transaction(fee_payer=authority).add(
        ixs.remove_market_product_ix(
            authority=authority,
            market_product_group=market_product_group,
            product_key=product_key,
            orderbook=orderbook,
            event_queue=event_queue,
            bids=bids,
            asks=asks,
        )
    )



import json
from random import random
import sys
import time

print('importing dexterity (it takes a while)...', end='', flush=True)
from dexteritysdk.dex.sdk_context import SDKContext, SDKTrader, MultiplaceOrder
from dexteritysdk.utils.aob import Side
print('success')
from solana.rpc.api import Client
from solders.keypair import Keypair
from solders.pubkey import Pubkey


CRANK_SLEEP_SECONDS = 5 # 0.400 * 31 # 400 ms per slot times 31 slots for confirmed
BTC_PRICE = 69000
ETH_PRICE = 4000
SOL_PRICE = 150


def load_keypair(fp):
    with open(fp, "r") as f:
        kp = json.load(f)
    return Keypair.from_bytes(kp)

async def main():
    network = "http://localhost:8899/"
    client = Client(network)

    print('loading keypairs...', end='', flush=True)
    keypair = load_keypair('../deploy_key.json')
    mpg_keypair = load_keypair('../target/deploy/mpg-keypair.json')
    print('success')

    print('connecting...', end='', flush=True)
    ctx = SDKContext.connect(
        client=client,
        market_product_group_key=mpg_keypair.pubkey(),
        payer=keypair,
        raise_on_error=True
    )
    print('success')

    print('creating new trg...', end='', flush=True)
    trader = ctx.register_trader(keypair)
    print('success (trg pubkey: {})'.format(trader.account))
    print('subscribing...', end='', flush=True)
    asyncio.create_task(trader.subscribe(ctx, 'ws://localhost:8900'))
    await asyncio.sleep(1)
    print('success')

    print('depositing 100k USDC...', end='', flush=True)
    trader.deposit(ctx, 100000)
    print('success')

    print('withdrawing 1k USDC...', end='', flush=True)
    trader.withdraw(ctx, 1000)
    print('success')

    print('checking balance...', end='', flush=True)
    trg, _ = trader.get_trader_risk_group()
    trg.cash_balance
    print('success... (cash balance: {})'.format(trg.cash_balance))

    print('checking balance...', end='', flush=True)
    trg, _ = trader.get_trader_risk_group()
    trg.cash_balance
    print('success... (cash balance: {})'.format(trg.cash_balance))

    print('checking mark prices')
    for product in ctx.products:
        mark_price = ctx.get_mark_price(product.key)
        index_price = ctx.get_index_price(product.key)
        oracle_minus_book_ewma = ctx.get_oracle_minus_book_ewma(product.key)
        print(f"{product.name}: mark: {mark_price}, index: {index_price}, spread ema: {oracle_minus_book_ewma}")

    tick = 0.1
    size = tick

    print('checking open orders...', end='', flush=True)
    open_orders = list(trader.open_orders(ctx))
    if len(open_orders) != 0:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected zero open orders but saw but saw {len(open_orders)} open orders")
    print('success')

    for product in ctx.products:
        if 'BTC' in product.name:
            price = BTC_PRICE
            size = 0.1
        elif 'ETH' in product.name:
            price = ETH_PRICE
            size = 0.1
        elif 'SOL' in product.name:
            price = SOL_PRICE
            size = 0.1
        elif 'BITCOIN' in product.name:
            price = 0.06
            size = 1
        elif 'OPOS' in product.name:
            price = 6
            size = 1
        print(f"multiplacing orders on {product.name}...", end='', flush=True)
        # TODO: actually figure out how to return a List[OrderSummary] from trader.multiplace
        order_summaries = trader.multiplace(ctx, product, [
            MultiplaceOrder(
                side=Side.ASK,
                size=size,
                price=price,
                client_order_id=0,
            ),
            MultiplaceOrder(
                side=Side.BID,
                size=size,
                price=price,
                client_order_id=1,
            ),
        ])
        print(order_summaries)
        print('success')

        # once the following TODO is done, we can run this code:
        # TODO: actually figure out how to return a List[OrderSummary] from trader.multiplace
        # print(f"replacing order on {product.name}...", end='', flush=True)
        # order_summary = trader.replace(
        #     ctx,
        #     product,
        #     order_summary.order_id,
        #     Side.ASK,
        #     size,
        #     price,
        # )
        # print('success')
        # 
        # print(f"cancelling order on {product.name}...", end='', flush=True)
        # trader.cancel(ctx, product, order_summary.order_id)
        # print('success')
    print('success')

    print('waiting a few seconds for the crank...', end='', flush=True)
    await asyncio.sleep(CRANK_SLEEP_SECONDS)
    print('done waiting')

    batch_size=2
    print(f"calling cancel_all_orders...", end='', flush=True)
    trader.cancel_all_orders(ctx)
    print('success')

    print('waiting a few seconds for the crank...', end='', flush=True)
    await asyncio.sleep(CRANK_SLEEP_SECONDS)
    print('done waiting')

    print('checking open orders...', end='', flush=True)
    open_orders = list(trader.open_orders(ctx))
    if len(open_orders) != 0:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected zero open orders but saw but saw {len(open_orders)} open orders")
    print('success')

    for product in ctx.products:
        if 'BTC' in product.name:
            price = BTC_PRICE
            size = 0.1
        elif 'ETH' in product.name:
            price = ETH_PRICE
            size = 0.1
        elif 'SOL' in product.name:
            price = SOL_PRICE
            size = 0.1
        elif 'BITCOIN' in product.name:
            price = 0.06
            size = 1
        elif 'OPOS' in product.name:
            price = 6
            size = 1
        client_order_id = int(random() * 1e6)
        print(f"placing order with client_order_id={client_order_id} on {product.name}...", end='', flush=True)
        trader.place_order(ctx, product, Side.ASK, size, price, client_order_id=client_order_id)
        print('success')

        print('checking open orders...', end='', flush=True)
        open_orders = list(trader.open_orders(ctx))
        if len(open_orders) != 1:
            print("printing open orders")
            for o in open_orders:
                print(o)
            raise RuntimeError(f"expected 1 open order but saw {len(open_orders)} open orders")
        if open_orders[0].client_order_id != client_order_id:
            print("printing open orders")
            for o in open_orders:
                print(o)
            raise RuntimeError(f"expected open_orders[0].client_order_id = {client_order_id} but saw {open_orders[0].client_order_id}")
        print('success')

        print(f"cancelling order with client_order_id={client_order_id} on {product.name}...", end='', flush=True)
        trader.cancel(ctx, product, 0, client_order_id=client_order_id)
        print('success')

        print('waiting a few seconds for the crank...', end='', flush=True)
        await asyncio.sleep(CRANK_SLEEP_SECONDS)
        print('done waiting')
        print('checking open orders...', end='', flush=True)
        open_orders = list(trader.open_orders(ctx))
        if len(open_orders) != 0:
            print("printing open orders")
            for o in open_orders:
                print(o)
            raise RuntimeError(f"expected zero open orders but saw but saw {len(open_orders)} open orders")
        print('success')

    print('success')

    print('waiting a few seconds for the crank...', end='', flush=True)
    await asyncio.sleep(CRANK_SLEEP_SECONDS)
    print('done waiting')
    print('checking open orders...', end='', flush=True)
    open_orders = list(trader.open_orders(ctx))
    if len(open_orders) != 0:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected zero open orders but saw but saw {len(open_orders)} open orders")
    print('success')

    for product in ctx.products:
        if 'BTC' in product.name:
            price = BTC_PRICE
            size = 0.1
        elif 'ETH' in product.name:
            price = ETH_PRICE
            size = 0.1
        elif 'SOL' in product.name:
            price = SOL_PRICE
            size = 0.1
        elif 'BITCOIN' in product.name:
            price = 0.06
            size = 1
        elif 'OPOS' in product.name:
            price = 6
            size = 1
        print(f"placing order on {product.name}...", end='', flush=True)
        order_summary = trader.place_order(ctx, product, Side.ASK, size, price)
        print('success')

    batch_size = 1
    print(f"calling cancel_all_orders...", end='', flush=True)
    trader.cancel_all_orders(ctx)
    print('success')

    batch_size = 2
    for product in ctx.products:
        if 'BTC' in product.name:
            price = BTC_PRICE
            size = 0.1
        elif 'ETH' in product.name:
            price = ETH_PRICE
            size = 0.1
        elif 'SOL' in product.name:
            price = SOL_PRICE
            size = 0.1
        elif 'BITCOIN' in product.name:
            price = 0.06
            size = 1
        elif 'OPOS' in product.name:
            price = 6
            size = 1
        for i in range(batch_size):
            print(f"placing order on {product.name}...", end='', flush=True)
            order_summary = trader.place_order(ctx, product, Side.ASK, size + i*tick, price)
            print('success')

    print(f"calling cancel_all_orders...", end='', flush=True)
    trader.cancel_all_orders(ctx)
    print('success')

    batch_size = 4
    for product in ctx.products:
        if 'BTC' in product.name:
            price = BTC_PRICE
            size = 0.1
        elif 'ETH' in product.name:
            price = ETH_PRICE
            size = 0.1
        elif 'SOL' in product.name:
            price = SOL_PRICE
            size = 0.1
        elif 'BITCOIN' in product.name:
            price = 0.06
            size = 1
        elif 'OPOS' in product.name:
            price = 6
            size = 1
        for i in range(batch_size):
            print(f"placing order on {product.name}...", end='', flush=True)
            order_summary = trader.place_order(ctx, product, Side.ASK, size + i*tick, price)
            print('success')

    print(f"calling cancel_all_orders...", end='', flush=True)
    trader.cancel_all_orders(ctx)
    print('success')

    batch_size = 8
    for product in ctx.products:
        if 'BTC' in product.name:
            price = BTC_PRICE
            size = 0.1
        elif 'ETH' in product.name:
            price = ETH_PRICE
            size = 0.1
        elif 'SOL' in product.name:
            price = SOL_PRICE
            size = 0.1
        elif 'BITCOIN' in product.name:
            price = 0.06
            size = 1
        elif 'OPOS' in product.name:
            price = 6
            size = 1
        for i in range(batch_size):
            print(f"placing order on {product.name}...", end='', flush=True)
            order_summary = trader.place_order(ctx, product, Side.ASK, size + i*tick, price)
            print('success')

    print(f"calling cancel_all_orders...", end='', flush=True)
    trader.cancel_all_orders(ctx)
    print('success')

    for product in ctx.products:
        if 'BTC' in product.name:
            price = BTC_PRICE
            size = 0.1
        elif 'ETH' in product.name:
            price = ETH_PRICE
            size = 0.1
        elif 'SOL' in product.name:
            price = SOL_PRICE
            size = 0.1
        elif 'BITCOIN' in product.name:
            price = 0.06
            size = 1
        elif 'OPOS' in product.name:
            price = 6
            size = 1
        print(f"placing order on {product.name}...", end='', flush=True)
        order_summary = trader.place_order(ctx, product, Side.ASK, size, price)
        print('success')

        print(f"replacing order on {product.name}...", end='', flush=True)
        order_summary = trader.replace(
            ctx,
            product,
            order_summary.order_id,
            Side.ASK,
            size,
            price,
        )
        print('success')

        print(f"cancelling order on {product.name}...", end='', flush=True)
        trader.cancel(ctx, product, order_summary.order_id)
        print('success')
    print('success')

    print('waiting a few seconds for the crank...', end='', flush=True)
    await asyncio.sleep(CRANK_SLEEP_SECONDS)
    print('done waiting')
    print('checking open orders...', end='', flush=True)
    open_orders = list(trader.open_orders(ctx))
    if len(open_orders) != 0:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected zero open orders but saw but saw {len(open_orders)} open orders")
    print('success')

if __name__ == "__main__":
    import asyncio
    asyncio.run(main())



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    U128,
    U16,
    U64,
    pod,
    U32,
)

# LOCK-END


# LOCK-BEGIN[class(OpenOrdersNode)]: DON'T MODIFY
@pod
class OpenOrdersNode:
    id: U128
    qty: U64
    client_id: U64
    prev: U16
    next: U16
    padding: U32
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.solmate.anchor import InstructionDiscriminant
from podite import (
    Enum,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[instruction_tag]: DON'T MODIFY
@pod
class InstructionTag(Enum[U64]):
    RESIZE_VARIANCE_CACHE = InstructionDiscriminant()
    VALIDATE_ACCOUNT_HEALTH = InstructionDiscriminant()
    VALIDATE_ACCOUNT_LIQUIDATION = InstructionDiscriminant()
    CREATE_RISK_STATE_ACCOUNT = InstructionDiscriminant()
    INITIALIZE_COVARIANCE_MATRIX = InstructionDiscriminant()
    INITIALIZE_MARK_PRICES = InstructionDiscriminant()
    CLOSE_MARK_PRICES = InstructionDiscriminant()
    UPDATE_MARK_PRICES = InstructionDiscriminant()
    REMOVE_MARKET_PRODUCT_INDEX_FROM_VARIANCE_CACHE = InstructionDiscriminant()
    # LOCK-END



import os

import solders.system_program
import solders.sysvar
import spl.token.constants
from solders.pubkey import Pubkey

from dexteritysdk.scripts import extract_program_ids

try:
    programs = extract_program_ids.get_program_to_id()
except:
    programs = {}

RENT_PROGRAM_ID = solders.sysvar.RENT
CLOCK_PROGRAM_ID = solders.sysvar.CLOCK
SYSTEM_PROGRAM_ID = solders.system_program.ID
SPL_TOKEN_PROGRAM_ID = spl.token.constants.TOKEN_PROGRAM_ID

DEX_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("DEX", programs.get("dex", "FUfpR31LmcP1VSbz5zDaM7nxnH55iBHkpwusgrnhaFjL"))
)
INSTRUMENTS_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("INSTRUMENTS", programs.get("instruments", "8981bZYszfz1FrFVx7gcUm61RfawMoAHnURuERRJKdkq"))
)
ORACLE_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("DUMMY_ORACLE", programs.get("dummy_oracle", "GLkqj95yBQHqb42A7Lf5bK33NBRnH4LhACVWajjyrgv4"))
)
STAKING_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("STAKING", programs.get("staking", "2jmux3fWV5zHirkEZCoSMEgTgdYZqkE9Qx2oQnxoHRgA"))
)
# NOOP_RISK_ENGINE_PROGRAM_ID = Pubkey.from_string(
#     os.environ.get("NOOP_RISK_ENGINE", programs.get("noop_risk_engine", ""))
# )
# SIMPLE_RISK_ENGINE_PROGRAM_ID = Pubkey.from_string(
#     os.environ.get("SIMPLE_RISK_ENGINE", "4HzqJ5fXcE5W1YTLa8m6M3cdSuGGjuBmDy9zNSTSnkJY")  # todo remove
# )
# ALPHA_RISK_ENGINE_PROGRAM_ID = Pubkey.from_string(
#     os.environ.get("ALPHA_RISK_ENGINE", programs.get("alpha_risk_engine", ""))
# )

# this is to get things going as this is imported in a few places but seems unused
ALPHA_RISK_ENGINE_PROGRAM_ID = Pubkey.default()

RISK_ENGINE_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("RISK_ENGINE", "92wdgEqyiDKrcbFHoBTg8HxMj932xweRCKaciGSW3uMr")
)
AOB_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("AGNOSTIC_ORDERBOOK",
                   programs.get("agnostic_orderbook", "DchhQ6g8LyRCM5mnao1MAg3g9twfqBbDmUWgpQpFfn1b"))
)
CONSTANT_FEES_MODEL_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("CONSTANT_FEES", programs.get("constant_fees", "5AZioCPiC7uZ4zRmkKSg5nsb2A98RhmW89a1pMwiDoeT"))
)
INCENTIVE_PROGRAM_ID = Pubkey.from_string(
    os.environ.get("INCENTIVE", programs.get("incentive", "5u8mLVnUSQNSbKdZPNGTfWHGwV5uJh9by5Fa6jb6BP6h"))
)

# todo: make this ~better~ -> work...
# set_pid_by_protocol_name("risk", ALPHA_RISK_ENGINE_PROGRAM_ID)
# set_pid_by_protocol_name("dex", DEX_PROGRAM_ID)
# set_pid_by_protocol_name("instruments", INSTRUMENTS_PROGRAM_ID)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.price_ewma import PriceEwma
from podite import (
    FixedLenArray,
    U64,
    U8,
    pod,
)
from solders.pubkey import Pubkey

# LOCK-END


# LOCK-BEGIN[class(ProductMetadata)]: DON'T MODIFY
@pod
class ProductMetadata:
    bump: U64
    product_key: Pubkey
    name: FixedLenArray[U8, 16]
    orderbook: Pubkey
    tick_size: "Fractional"
    base_decimals: U64
    price_offset: "Fractional"
    notional_traded_volume: "Fractional"
    prices: PriceEwma
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    AutoTagType,
    Enum,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(DexError)]: DON'T MODIFY
@pod
class DexError(Enum[AutoTagType]):
    CONTRACT_IS_EXPIRED = None
    CONTRACT_IS_NOT_EXPIRED = None
    INVALID_SYSTEM_PROGRAM_ACCOUNT = None
    INVALID_AOB_PROGRAM_ACCOUNT = None
    INVALID_STATE_ACCOUNT_OWNER = None
    INVALID_ORDER_INDEX = None
    USER_ACCOUNT_FULL = None
    TRANSACTION_ABORTED = None
    MISSING_USER_ACCOUNT = None
    ORDER_NOT_FOUND = None
    NO_OP = None
    OUTOF_FUNDS = None
    USER_ACCOUNT_STILL_ACTIVE = None
    MARKET_STILL_ACTIVE = None
    INVALID_MARKET_SIGNER_ACCOUNT = None
    INVALID_ORDERBOOK_ACCOUNT = None
    INVALID_MARKET_ADMIN_ACCOUNT = None
    INVALID_BASE_VAULT_ACCOUNT = None
    INVALID_QUOTE_VAULT_ACCOUNT = None
    FULL_MARKET_PRODUCT_GROUP = None
    MISSING_MARKET_PRODUCT = None
    INVALID_WITHDRAWAL_AMOUNT = None
    INVALID_TAKER_TRADER = None
    FUNDS_ERROR = None
    INACTIVE_PRODUCT_ERROR = None
    TOO_MANY_OPEN_ORDERS_ERROR = None
    NO_MORE_OPEN_ORDERS_ERROR = None
    NON_ZERO_PRICE_TICK_EXPONENT_ERROR = None
    DUPLICATE_PRODUCT_NAME_ERROR = None
    INVALID_RISK_RESPONSE_ERROR = None
    INVALID_ACCOUNT_HEALTH_ERROR = None
    ORDERBOOK_IS_EMPTY_ERROR = None
    COMBOS_NOT_REMOVED = None
    ACCOUNT_NOT_LIQUIDABLE = None
    FUNDING_PRECISION_ERROR = None
    PRODUCT_DECIMAL_PRECISION_ERROR = None
    PRODUCT_NOT_OUTRIGHT = None
    PRODUCT_NOT_COMBO = None
    INVALID_SOCIAL_LOSS_CALCULATION = None
    PRODUCT_INDEX_MISMATCH = None
    INVALID_ORDER_I_D = None
    INVALID_BYTES_FOR_ZERO_COPY_DESERIALIZATION = None
    INCORRECT_PRINT_TRADE_SIZE = None
    INCORRECT_PRINT_TRADE_PRICE = None
    INCORRECT_PRINT_TRADE_SIDE = None
    INCORRECT_PRINT_TRADE_OPERATOR_CREATOR_FEES = None
    INCORRECT_PRINT_TRADE_OPERATOR_COUNTERPARTY_FEES = None
    INVALID_PRINT_TRADE_OPERATOR_FEES = None
    DEPOSIT_LIMIT_EXCEEDED = None
    WITHDRAW_LIMIT_EXCEEDED = None
    NEGATIVE_DEPOSIT_LIMIT = None
    NEGATIVE_WITHDRAW_LIMIT = None
    DEPOSIT_DENIED_INSUFFICIENT_BALANCE_ON_WHITELIST_ATA_TOKEN = None
    DEPOSIT_DECLINED_UNFROZEN_WHITELIST_ATA_TOKEN = None
    DEPOSIT_DECLINED_NON_EXISTENT_WHITELIST_ATA_TOKEN_ON_TRADER_RISK_GROUP = None
    INVALID_PRODUCT_STATUS_IN_UPDATE_FUNDING = None
    CONTRACT_IS_NOT_EXPIRING = None
    CONTRACT_HAS_NON_ZERO_OPEN_INTEREST = None
    CONTRACT_HAS_NON_ZERO_OPEN_INTEREST_OR_RISK_STATE_ACCOUNTS = None
    CONTRACT_IS_ACTIVE = None
    FAILED_TO_GET_ORDER_QUANTITY = None
    SELF_TRADE_BEHAVIOR_DECREMENT_TAKE_IS_DISALLOWED = None
    PRICE_BAND_VIOLATION = None
    UNEXPECTED_IMBALANCED_OPEN_INTEREST = None
    MAXIMUM_OPEN_INTEREST_EXCEEDED = None
    MARKET_PRODUCT_GROUP_KILLSWITCH_IS_ON = None
    INVALID_FUTURE_EXPIRY = None
    MAX_REFERRER_FEE_BPS_EXCEEDED = None
    PRINT_TRADE_OPERATOR_DID_NOT_SIGN = None
    PRINT_TRADE_INVALID_PRODUCTS_LENGTH = None
    CONTRACT_IS_NOT_ACTIVE = None
    PRINT_TRADE_INVALID_NUM_PRODUCTS = None
    PRINT_TRADE_PRODUCT_MISMATCH = None
    INSUFFICIENT_LOCKED_COLLATERAL = None
    ORACLE_NOT_WHITELISTED = None
    NEGATIVE_LIQUIDATION_PRICE_DOES_NOT_EQUAL_SOCIAL_LOSS = None
    NEGATIVE_SOCIAL_LOSS = None
    SOCIAL_LOSS_NEGATIVE_TOTAL_APPLICABLE_SHARES = None
    MARKET_PRODUCT_GROUP_ADMIN_MODE_IS_ON = None
    MATCH_LIMIT_ZERO = None
    # LOCK-END

    @classmethod
    def _to_bytes_partial(cls, buffer, obj, **kwargs):
        # to modify packing, change this method
        return super()._to_bytes_partial(buffer, obj, **kwargs)

    @classmethod
    def _from_bytes_partial(cls, buffer, **kwargs):
        # to modify unpacking, change this method
        return super()._from_bytes_partial(buffer, **kwargs)

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from dataclasses import dataclass
from enum import Enum
from typing import Union

from anchorpy import Event
from solders.pubkey import Pubkey

from dexteritysdk.utils.aob.state import Side
from decimal import *

class CompletedReason(Enum):
    Cancelled = 0
    Filled = 1
    Booted = 2
    SelfTradeCancel = 3
    PostNotAllowed = 4
    MatchLimitExhausted = 5
    PostOnly = 6
    Expired = 7


def decimal_from_int(num) -> Decimal:
    return Decimal(num).scaleb(-8)

def bps_from_int(num) -> Decimal:
    return Decimal(num).scaleb(-4)


@dataclass
class OrderPlacedEvent:
    market_product_group: Pubkey
    trader_risk_group: Pubkey
    product: str
    order_id: int
    client_order_id: int
    side: Side
    price: Decimal
    size: Decimal
    filled_size: Decimal
    posted_size: Decimal
    referrer_trader_risk_group: Pubkey
    referrer_fee_bps: Decimal

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            trader_risk_group=event.data.trader_risk_group,
            product=bytes(event.data.product).decode("utf-8").strip(),
            order_id=event.data.order_id,
            client_order_id=event.data.client_order_id,
            side=Side(event.data.side.index),
            price=decimal_from_int(event.data.price),
            size=decimal_from_int(event.data.size),
            filled_size=decimal_from_int(event.data.filled_size),
            posted_size=decimal_from_int(event.data.posted_size),
            referrer_trader_risk_group=event.data.referrer_trader_risk_group,
            referrer_fee_bps=bps_from_int(event.data.referrer_fee_bps),
        )


@dataclass
class OrderFillEvent:
    market_product_group: Pubkey
    product: str
    maker_trader_risk_group: Pubkey
    maker_order_id: int
    maker_order_nonce: int
    maker_client_order_id: int
    maker_fee: Decimal
    taker_trader_risk_group: Pubkey
    taker_order_id: int
    taker_order_nonce: int
    taker_client_order_id: int
    taker_fee: Decimal
    taker_side: Side
    price: Decimal
    base_size: Decimal
    quote_size: Decimal

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            product=bytes(event.data.product).decode("utf-8").strip(),
            maker_trader_risk_group=event.data.maker_trader_risk_group,
            maker_order_id=event.data.maker_order_id,
            maker_order_nonce=event.data.maker_order_nonce,
            maker_client_order_id=event.data.maker_client_order_id,
            maker_fee=decimal_from_int(event.data.maker_fee),
            taker_trader_risk_group=event.data.taker_trader_risk_group,
            taker_order_id=event.data.taker_order_id,
            taker_order_nonce=event.data.taker_order_nonce,
            taker_client_order_id=event.data.taker_client_order_id,
            taker_fee=decimal_from_int(event.data.taker_fee),
            taker_side=Side(event.data.taker_side.index),
            price=decimal_from_int(event.data.price),
            base_size=decimal_from_int(event.data.base_size),
            quote_size=decimal_from_int(event.data.quote_size),
        )


@dataclass
class OrderCompletedEvent:
    market_product_group: Pubkey
    trader_risk_group: Pubkey
    product: str
    order_id: int
    client_order_id: int
    side: Side
    size: Decimal
    reason: CompletedReason

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            trader_risk_group=event.data.trader_risk_group,
            product=bytes(event.data.product).decode("utf-8").strip(),
            order_id=event.data.order_id,
            client_order_id=event.data.client_order_id,
            side=Side(event.data.side.index),
            size=decimal_from_int(event.data.size),
            reason=CompletedReason(event.data.reason.index)
        )


@dataclass
class OrderCancelledEvent:
    market_product_group: Pubkey
    trader_risk_group: Pubkey
    product: str
    order_id: int
    client_order_id: int
    side: Side
    size: Decimal

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            trader_risk_group=event.data.trader_risk_group,
            product=bytes(event.data.product).decode("utf-8").strip(),
            order_id=event.data.order_id,
            client_order_id=event.data.client_order_id,
            side=Side(event.data.side.index),
            size=decimal_from_int(event.data.size),
        )


@dataclass
class TraderApplyFundingEvent:
    market_product_group: Pubkey
    trader_risk_group: Pubkey
    product: str
    funding_amount: Decimal
    social_loss_amount: Decimal
    remaining_cash_balance: Decimal

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            trader_risk_group=event.data.trader_risk_group,
            product=bytes(event.data.product).decode("utf-8").strip(),
            funding_amount=decimal_from_int(event.data.funding_amount),
            social_loss_amount=decimal_from_int(event.data.social_loss_amount),
            remaining_cash_balance=decimal_from_int(event.data.remaining_cash_balance),
        )


@dataclass
class TraderDepositEvent:
    market_product_group: Pubkey
    trader_risk_group: Pubkey
    qty: Decimal

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            trader_risk_group=event.data.trader_risk_group,
            qty=decimal_from_int(event.data.qty),
        )


@dataclass
class TraderWithdrawalEvent:
    market_product_group: Pubkey
    trader_risk_group: Pubkey
    qty: Decimal

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            trader_risk_group=event.data.trader_risk_group,
            qty=decimal_from_int(event.data.qty),
        )


@dataclass
class TraderLiquidationEvent:
    market_product_group: Pubkey
    liquidator_trader_risk_group: Pubkey
    liquidatee_trader_risk_group: Pubkey
    liquidator_cash_balance: Decimal
    liquidatee_cash_balance: Decimal
    liquidation_price: Decimal

    @classmethod
    def from_event(cls, event: Event):
        assert event.name == cls.__name__
        return cls(
            market_product_group=event.data.market_product_group,
            liquidator_trader_risk_group=event.data.liquidator_trader_risk_group,
            liquidatee_trader_risk_group=event.data.liquidatee_trader_risk_group,
            liquidator_cash_balance=decimal_from_int(event.data.liquidator_cash_balance),
            liquidatee_cash_balance=decimal_from_int(event.data.liquidatee_cash_balance),
            liquidation_price=decimal_from_int(event.data.liquidation_price),
        )


DexEvent = Union[OrderPlacedEvent, OrderCompletedEvent, OrderFillEvent, OrderCancelledEvent, TraderLiquidationEvent,
                 TraderWithdrawalEvent, TraderDepositEvent, TraderApplyFundingEvent]



from __future__ import annotations

from typing import Iterable, List, Union

from dexteritysdk.pyserum.layouts.slab import SlabType
from dexteritysdk.pyserum.serum_types import OrderInfo, Order

from dexteritysdk.pyserum.enums import Side
from dexteritysdk.pyserum.structs.slab import Slab, SlabInnerNode, SlabLeafNode
from dexteritysdk.codegen.dex.types import Fractional


class OrderBook:
    """Represents an order book."""

    _is_bids: bool
    _slab: Slab
    _price_offset: Fractional
    _tick_size: Fractional

    def __init__(self, slab: Slab, decimals: int, price_offset: Fractional, tick_size: Fractional) -> None:
        self._is_bids = (slab._header.tag == SlabType.BIDS)
        self._slab = slab
        self._decimals = decimals
        self._price_offset = price_offset
        self._tick_size = tick_size

    @staticmethod
    def get_price_from_key(key: int, tick_size: Fractional, price_offset: Fractional) -> float:
        return ((key >> 96) * tick_size - price_offset).value

    @staticmethod
    def key_is_bid(key: int) -> bool:
        return (key >> 63) & 1 != 0

    def __get_price_from_slab(self, node: Union[SlabInnerNode, SlabLeafNode]) -> float:
        """Get price from a slab node key.

        The key is constructed as the (price << 64) + (seq_no if ask_order else !seq_no).
        """
        return OrderBook.get_price_from_key(node.key, self._tick_size, self._price_offset)

    @staticmethod
    def from_bytes(buffer: bytes, decimals: int, price_offset: Fractional, tick_size: Fractional) -> OrderBook:
        """Decode the given buffer into an order book."""
        slab = Slab.from_bytes(buffer[:])
        return OrderBook(slab, decimals, price_offset, tick_size)

    def get_l2(self, depth: int) -> List[OrderInfo]:
        """Get the Level 2 market information."""
        descending = self._is_bids
        # The first element of the inner list is price, the second is quantity.
        levels: List[List[float]] = []
        for node in self._slab.items(descending):
            price = self.__get_price_from_slab(node)
            if len(levels) > 0 and levels[len(levels) - 1][0] == price:
                levels[len(levels) - 1][1] += node.quantity
            elif len(levels) == depth:
                break
            else:
                levels.append([price, node.quantity])
        return [
            OrderInfo(
                price=price_lots,
                size=size_lots / (10 ** self._decimals),
            )
            for price_lots, size_lots in levels
        ]

    def __iter__(self) -> Iterable[Order]:
        return self.orders()

    def orders(self) -> Iterable[Order]:
        for node in self._slab.items():
            price = self.__get_price_from_slab(node)

            yield Order(
                order_id=node.key,
                client_id=node.client_order_id,
                info=OrderInfo(
                    price=price,
                    size=node.quantity / (10 ** self._decimals),
                    trader=node.trader,
                    expiration_slot=node.expiration_slot,
                ),
                side=Side.BUY if self._is_bids else Side.SELL,
            )



from typing import List, Iterable

from dexteritysdk.pyserum.structs.event_queue import EventQueueHeader, Register, Event, FillEvent, OutEvent
from dexteritysdk.pyserum.layouts.event_queue import EVENT_QUEUE_HEADER_LAYOUT, REGISTER, EVENT_NODE_LAYOUT, EventType


class EventQueue:
    _header: EventQueueHeader
    _register: Register
    _events: List[Event]

    def __init__(self, header: EventQueueHeader, register: Register, events: List[Event]):
        self._header = header
        self._register = register
        self._events = events

    def __iter__(self) -> Iterable[Event]:
        return self.items()

    def items(self) -> Iterable[Event]:
        for e in self._events:
            yield e

    @staticmethod
    def from_bytes(buffer: bytes) -> "EventQueue":
        header_size = EVENT_QUEUE_HEADER_LAYOUT.sizeof()
        register_size = REGISTER.sizeof()

        header = EVENT_QUEUE_HEADER_LAYOUT.parse(buffer[:header_size])
        register = REGISTER.parse(buffer[header_size:header_size+register_size])
        events = EventQueue.__build_events(buffer, header_size + register_size, header.head, header.count,
                                           header.event_size)

        return EventQueue(
            EventQueueHeader.from_construct(header),
            Register.from_construct(register),
            events
        )

    @staticmethod
    def __build_events(buffer: bytes, header_offset: int, head: int, count: int, event_size: int) -> List[Event]:
        result = []
        for i in range(count):
            node_offset = header_offset + ((i * event_size + head) % (len(buffer) - header_offset))
            event = EVENT_NODE_LAYOUT.parse(buffer[node_offset:node_offset + event_size])
            if event.tag == EventType.FILL:
                result.append(FillEvent.from_construct(event.data))
            else:
                result.append(OutEvent.from_construct(event.data))
        return result



#!/usr/bin/env python3

import argparse
import os.path

from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana import system_program
import base58

from dexteritysdk.dex.sdk_context import SDKContext
from dexteritysdk.program_ids import programs
from dexteritysdk.utils.solana import Client, send_instructions, Context, explore
from dexteritysdk.codegen.dex.types import Fractional

DEVNET_URL = "https://api.devnet.solana.com"
MAINNET_URL = "https://api.mainnet-beta.solana.com"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--market_product_group", required=True, type=Pubkey)
    parser.add_argument("-n", "--network", default=DEVNET_URL)
    args = parser.parse_args()
    mpg_key = Pubkey.from_string(args.market_product_group)
    print(str(base58.b58encode(bytes(mpg_key))))

    payer_keypair_b58 = os.environ["PRIVATE_KEY"]
    payer_keypair_bytes = base58.b58decode(payer_keypair_b58)
    payer_keypair = Keypair.from_secret_key(payer_keypair_bytes)

    client = Client(args.network)
    Context.set_raise_on_error(True)

    sdk = SDKContext.connect(client, market_product_group_key=mpg_key, payer=payer_keypair, **programs, raise_on_error=True)
    trader = create_test_trader(sdk, fund_from_mint=False)


def create_test_trader(sdk: SDKContext, fund_from_mint=True):
    from spl.token import instructions
    from spl.token.constants import TOKEN_PROGRAM_ID

    keypair = Keypair()
    Context.add_signers((keypair, "test_trader"))
    # transfer lamports since 'keypair' will be payer
    send_instructions(system_program.transfer(system_program.TransferParams(
        from_pubkey=sdk.payer.pubkey(),
        to_pubkey=keypair.pubkey(),
        lamports=1_000_000_000,  # 1 SOL
    )))

    # create and mint to ata to fund trader
    wallet = instructions.get_associated_token_address(keypair.pubkey(), sdk.vault_mint)
    ata_ix = instructions.create_associated_token_account(sdk.payer.pubkey(), keypair.pubkey(), sdk.vault_mint)
    ixs = [ata_ix]
    if fund_from_mint:
        mint_ix = instructions.mint_to(instructions.MintToParams(
            program_id=TOKEN_PROGRAM_ID,
            mint=sdk.vault_mint,
            dest=wallet,
            mint_authority=sdk.payer.pubkey(),
            amount=200_000_000,
        ))
        ixs.append(mint_ix)
    else:
        print(f"Please get the MPG admin to mint vault tokens to the trader. The vault token mint is {sdk.vault_mint}")

    send_instructions(*ixs)
    trader = sdk.register_trader(keypair=keypair, wallet=wallet)
    print(f"Trader pubkey is {str(keypair.pubkey())}")
    print(f"Trader keypair is {str(base58.b58encode(keypair.secret_key))}")
    print(f"Trader risk group is {str(trader.account)}")
    print(f"Trader vault token ATA is {str(wallet)}")

    # deposit minted tokens to dex
    if fund_from_mint:
        print("Depositing vault tokens to dex")
        trader.deposit(sdk, Fractional(100_000_000, sdk.decimals))

    return trader


if __name__ == '__main__':
    main()



import hashlib
from typing import Optional, Tuple

from podite import U8
from solders.pubkey import Pubkey

from dexteritysdk import program_ids as pids
from dexteritysdk.utils.aob.state import MarketState

DEFAULT_ORDERBOOK_SIZE = MarketState.calc_size()
DEFAULT_EVENT_QUEUE_SIZE = 100_000
DEFAULT_ASKS_SIZE = 100_000
DEFAULT_BIDS_SIZE = 100_000
DEFAULT_TICK_SIZE = 0.1
DEFAULT_DECIMALS = 6
DEFAULT_MATCH_LIMIT = 50
DEFAULT_OFFSET = 0
OUT_REGISTER_RISK_SIZE = 440  # fixed size taken from rust

PROGRAMS_SEED = f"{pids.AOB_PROGRAM_ID}:{pids.DEX_PROGRAM_ID}"
MARKET_PRODUCT_GROUP_SEED_LAYOUT = f"prod_grp:{PROGRAMS_SEED}:{{seed}}"
ORDERBOOK_SEED_LAYOUT = f"prod:ob:{PROGRAMS_SEED}:{{group}}:{{key}}"
EVENT_QUEUE_SEED_LAYOUT = f"prod:eq:{PROGRAMS_SEED}:{{group}}:{{key}}"
BIDS_SEED_LAYOUT = f"prod:bid:{PROGRAMS_SEED}:{{group}}:{{key}}"
ASKS_SEED_LAYOUT = f"prod:ask:{PROGRAMS_SEED}:{{group}}:{{key}}"
TRADER_RISK_GROUP_SEED_LAYOUT = f"trdr_grp:{PROGRAMS_SEED}:{{market_product_group}}"
MINT_SEED_LAYOUT = f"mint:{PROGRAMS_SEED}:{{seed}}"
OUT_REGISTER_RISK_LAYOUT = f"out_register_risk:{PROGRAMS_SEED}:{{group}}"
FEE_REGISTER_LAYOUT = f"fee:{PROGRAMS_SEED}:{{group}}"
RISK_CONFIG_LAYOUT = f"risk_config:{PROGRAMS_SEED}:{{group}}"
IN_REGISTER_RISK_LAYOUT = f"in_register_risk:{PROGRAMS_SEED}:{{group}}"


def crush(string):
    return hashlib.md5(string.encode("UTF-8")).hexdigest()


def get_market_product_group_addr(
        authority: Pubkey,
        seed: str,
):
    return Pubkey.create_with_seed(
        from_pubkey=authority,
        seed=crush(MARKET_PRODUCT_GROUP_SEED_LAYOUT.format(seed=seed)),
        program_id=pids.DEX_PROGRAM_ID,
    )


def get_risk_signer(mpg: Pubkey):
    return Pubkey.find_program_address([bytes(mpg)], pids.DEX_PROGRAM_ID)[0]


def get_risk_register_addr(
        authority: Pubkey,
        group: Pubkey,
        program_id: Pubkey,
        layout_str: str,
):
    return Pubkey.create_with_seed(
        from_pubkey=authority,
        seed=crush(layout_str.format(group=group)),
        program_id=program_id,
    )


def get_fee_register_addr(
        authority: Pubkey,
        group: Pubkey,
        program_id: Pubkey,
        layout_str: str,
):
    return Pubkey.create_with_seed(
        from_pubkey=authority,
        seed=crush(layout_str.format(group=group)),
        program_id=program_id,
    )


def get_trader_fee_state_acct(
        trader_risk_group: Pubkey,
        market_product_group: Pubkey,
        fee_model_config: Pubkey,
        program_id: Pubkey = pids.CONSTANT_FEES_MODEL_PROGRAM_ID,
) -> Pubkey:
    key, _ = Pubkey.find_program_address(
        seeds=[
            bytes(market_product_group),
            bytes(trader_risk_group),
            bytes(fee_model_config),
        ],
        program_id=program_id,
    )
    return key


def get_market_product_group_vault(mpg_key: Pubkey):
    return Pubkey.find_program_address(
        [b"market_vault", bytes(mpg_key)], pids.DEX_PROGRAM_ID
    )[0]


def get_market_signer(product_key: Pubkey, dex_pid: Pubkey = pids.DEX_PROGRAM_ID) -> Pubkey:
    return Pubkey.find_program_address([bytes(product_key)], dex_pid)[0]


def get_risk_model_configuration_addr(
        market_product_group_key: Pubkey,
        program_id: Optional[Pubkey] = None,
) -> Pubkey:
    if program_id is None:
        program_id = pids.CONSTANT_FEES_MODEL_PROGRAM_ID
    key, _ = Pubkey.find_program_address(
        seeds=[
            b"risk_model_config_acct",
            bytes(market_product_group_key),
        ],
        program_id=program_id,
    )
    return key


def get_fee_model_configuration_addr(
        market_product_group_key: Pubkey,
        program_id: Optional[Pubkey] = None,
) -> Pubkey:
    if program_id is None:
        program_id = pids.CONSTANT_FEES_MODEL_PROGRAM_ID
    key, _ = Pubkey.find_program_address(
        seeds=[
            b"fee_model_config_acct",
            bytes(market_product_group_key),
        ],
        program_id=program_id,
    )
    return key


def get_trader_risk_group_addr(
        trader: Pubkey,
        market_product_group: Pubkey,
):
    return Pubkey.create_with_seed(
        from_pubkey=trader,
        seed=crush(
            TRADER_RISK_GROUP_SEED_LAYOUT.format(
                market_product_group=market_product_group
            )
        ),
        program_id=pids.DEX_PROGRAM_ID,
    )



from ctypes import Union
from solders.instruction import AccountMeta, Instruction
from typing import Optional

from solders.pubkey import Pubkey
from solana.transaction import Transaction
import dexteritysdk.program_ids as pids
from podite import pod, U64
from dexteritysdk.utils.solana import (
    actionify
)
from dexteritysdk.dex.actions import RISK_CONFIG_LAYOUT
from dexteritysdk.dex.addrs import get_risk_register_addr

@pod
class Params:
    instr: U64

def _post_init_risk_config(resp):
    if resp is None:
        return None, None
    addr = resp.instructions[0]["accounts"][1]
    exists = False

    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify(post_process=_post_init_risk_config)
def initialize_risk_config_acct(
    admin: Pubkey,
    market_product_group: Pubkey,
    program_id: Optional[Pubkey] = pids.ALPHA_RISK_ENGINE_PROGRAM_ID,
    risk_model_config_acct: Optional[Pubkey] = None,
    layout_str: Optional[str] = RISK_CONFIG_LAYOUT,
):
    if program_id == pids.ALPHA_RISK_ENGINE_PROGRAM_ID: 
        return None

    if risk_model_config_acct is None:
        risk_model_config_acct = get_risk_register_addr(admin, market_product_group, program_id, layout_str)

    keys = [
        AccountMeta(pubkey=market_product_group, is_signer=False, is_writable=False),
        AccountMeta(pubkey=admin, is_signer=True, is_writable=False),
        AccountMeta(pubkey=risk_model_config_acct, is_signer=False, is_writable=True),        
    ]

    params = Params(
        instr=4,
    )

    return Transaction(fee_payer=admin).add(
        Instruction(
                accounts=keys,
                program_id=program_id,
                data=Params.to_bytes(params),
            ),
        )




# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.alpha_risk_engine.types.health import Health
from podite import (
    Enum,
    U64,
    pod,
)
from dexteritysdk.solmate.anchor import AccountDiscriminant

# LOCK-END


# LOCK-BEGIN[accounts]: DON'T MODIFY
@pod
class Accounts(Enum[U64]):
    HEALTH = AccountDiscriminant(field=Health)
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import os
from typing import Dict, Iterable, Set, Union, Callable, Literal, Optional

from podite import Vec
from solders.pubkey import Pubkey

from dexteritysdk.solmate.utils import camel_to_snake, pascal_to_snake, snake_to_pascal
from .editor import CodeEditor
from .idl import (
    Idl,
    IdlTypeDefinition,
    IdlTypeDefinitionTy,
    IdlType,
    EnumFields,
    IdlAccountItem,
    IdlInstruction,
)


class InstructionCodeGen:
    codegen: "CodeGen"

    def __init__(self, codegen: "CodeGen"):
        self.codegen = codegen

    @staticmethod
    def _flatten_accounts(accounts: Vec[IdlAccountItem], name_prefix=""):
        flat = []
        for account in accounts:
            if int(account) == int(IdlAccountItem.IDL_ACCOUNT):
                flat.append((account, name_prefix))
            else:
                prefix = f"{name_prefix}{account.field.name}_"
                flat += InstructionCodeGen._flatten_accounts(account.field.accounts, prefix)

        return flat

    def generate_instruction(self, module_editor: CodeEditor, instr: IdlInstruction):
        """
        Generates python files for constructing and serializing each instruction in the idl
        """
        codegen = self.codegen
        instr_name = camel_to_snake(instr.name)
        instr_accounts = self._flatten_accounts(instr.accounts)

        editor = codegen.get_editor(f"{codegen.root_module}.instructions.{instr_name}")
        editor.add_from_import("solders.instruction", "AccountMeta")
        editor.add_from_import("solders.pubkey", "Pubkey")
        editor.add_from_import("dataclasses", "dataclass")
        editor.add_from_import("podite", "BYTES_CATALOG")
        editor.add_from_import("typing", "Optional")
        editor.add_from_import("typing", "List")

        # generating instruction account metas
        module_editor.add_from_import(
            f".{instr_name}", f"{snake_to_pascal(instr_name)}Ix"
        )
        module_editor.add_from_import(f".{instr_name}", f"{instr_name}")
        code = [
            "@dataclass\n" f"class {snake_to_pascal(instr_name)}Ix:\n",
            "    program_id: Pubkey\n",
            "\n",
            "    # account metas\n",
        ]
        for account, prefix in instr_accounts:
            account_name = camel_to_snake(prefix + account.field.name)
            account_type = (
                "Optional[AccountMeta]" if account.field.is_optional else "AccountMeta"
            )
            code.append(f"    {account_name}: {account_type}\n")

        code.append(f"    remaining_accounts: Optional[List[AccountMeta]]\n")

        # generating instruction data fields
        if len(instr.args) > 0:
            code += [
                "\n",
                "    # data fields\n",
            ]
        for arg in instr.args:
            arg_type = codegen.get_type_as_string(arg.type, editor, within_types=False)
            code.append(f"    {arg.py_name}: {arg_type}\n")

        # generating to_instruction() function
        code.append("\n")
        code.append("    def to_instruction(self):\n")
        code.append("        keys = []\n")
        for account, prefix in instr_accounts:
            account_name = camel_to_snake(prefix + account.field.name)
            if account.field.is_optional:
                code.append(f"        if self.{account_name} is not None:\n")
                code.append(f"            keys.append(self.{account_name})\n")
            else:
                code.append(f"        keys.append(self.{account_name})\n")

        code.append(f"        if self.remaining_accounts is not None:\n")
        code.append(f"            keys.extend(self.remaining_accounts)\n")
        code.append("\n")

        # generating data
        editor.add_from_import("io", "BytesIO")
        code.append("        buffer = BytesIO()\n")
        editor.add_from_import(".instruction_tag", "InstructionTag")
        instr_tag_name = camel_to_snake(instr.name).upper()
        code.append(
            f"        buffer.write(InstructionTag.to_bytes(InstructionTag.{instr_tag_name}))\n"
        )
        for arg in instr.args:
            arg_type = codegen.get_type_as_string(arg.type, editor, within_types=False)
            code.append(
                f"        buffer.write(BYTES_CATALOG.pack({arg_type}, self.{arg.py_name}))\n"
            )
        code.append("\n")

        editor.add_from_import("solders.instruction", "Instruction")
        code.append("        return Instruction(\n")
        code.append("            accounts=keys,\n")
        code.append(f"            program_id=self.program_id,\n")
        code.append("            data=buffer.getvalue(),\n")
        code.append("        )\n")
        code.append("\n")
        if editor.set_with_lock(f"ix_cls({instr_name})", code):
            editor.add_lines("\n", "\n")

        # generating helper function's code
        code = [f"def {instr_name}(\n"]

        # arguments
        args_without_default = []
        args_with_default = []
        editor.add_from_import("typing", "Union")
        meta_type = "Union[str, Pubkey, AccountMeta]"
        for account, prefix in instr_accounts:
            account_name = camel_to_snake(prefix + account.field.name)
            account_type = (
                f"Optional[{meta_type}]" if account.field.is_optional else meta_type
            )
            default_account = codegen.get_account_address(editor, account_name)
            if default_account is not None:
                args_with_default.append((account_name, account_type, default_account))
            elif account.field.is_optional:
                args_with_default.append((account_name, account_type, "None"))
            else:
                args_without_default.append((account_name, account_type))

        for arg in instr.args:
            arg_type = codegen.get_type_as_string(arg.type, editor, within_types=False)
            args_without_default.append((arg.py_name, arg_type))

        args_with_default.append(
            ("remaining_accounts", "Optional[List[AccountMeta]]", "None")
        )
        args_with_default.append(("program_id", "Optional[Pubkey]", "None"))

        for arg_name, arg_type in args_without_default:
            code.append(f"    {arg_name}: {arg_type},\n")

        for arg_name, arg_type, arg_val in args_with_default:
            code.append(f"    {arg_name}: {arg_type} = {arg_val},\n")

        code.append(f"):\n")

        # helper's body
        code.append("    if program_id is None:\n")
        code.append(
            f"        program_id = {codegen.get_account_address(editor, 'program_id')}\n"
        )
        code.append("\n")

        editor.add_from_import("dexteritysdk.solmate.utils", "to_account_meta")
        for account, prefix in instr_accounts:
            account_name = camel_to_snake(prefix + account.field.name)
            code.append(f"    if isinstance({account_name}, (str, Pubkey)):\n")
            code.append(f"        {account_name} = to_account_meta(\n")
            code.append(f"            {account_name},\n")
            code.append(f"            is_signer={account.field.is_signer},\n")
            code.append(f"            is_writable={account.field.is_mut},\n")
            code.append(f"        )\n")
        code.append("\n")

        # creating the instruction and returning it as a regular instruction
        code.append(f"    return {snake_to_pascal(instr_name)}Ix(\n")
        code.append(f"        program_id=program_id,\n")
        for account, prefix in instr_accounts:
            account_name = camel_to_snake(prefix + account.field.name)
            code.append(f"        {account_name}={account_name},\n")
        code.append(f"        remaining_accounts=remaining_accounts,\n")
        for arg in instr.args:
            code.append(f"        {arg.py_name}={arg.py_name},\n")
        code.append(f"    ).to_instruction()\n")
        code.append(f"\n")

        # writing instruction class
        editor.set_with_lock(f"ix_fn({instr_name})", code)

    def generate(self):
        codegen = self.codegen
        if not codegen.idl.instructions:
            return

        module_editor = codegen.get_editor(
            f"{codegen.root_module}.instructions", is_file=False
        )
        for instr in codegen.idl.instructions:
            self.generate_instruction(module_editor, instr)

        # generating InstructionTag class
        instr_tag_editor = codegen.get_editor(
            f"{codegen.root_module}.instructions.instruction_tag"
        )
        instr_tag_editor.add_from_import("podite", "Enum")
        if codegen.instr_tag_values == "anchor":
            tag_type = "U64"
            variant_type = "InstructionDiscriminant"
            instr_tag_editor.add_from_import(
                "dexteritysdk.solmate.anchor", "InstructionDiscriminant"
            )
        else:
            tag_type = codegen.instr_tag_values.split(":")[1]
            variant_type = "Variant"
            instr_tag_editor.add_from_import("pod", "Variant")

        instr_tag_editor.add_from_import("podite", "pod")
        instr_tag_editor.add_from_import("podite", tag_type)
        instr_tag_code = [
            "@pod\n",
            f"class InstructionTag(Enum[{tag_type}]):\n",
        ]
        for instr in codegen.idl.instructions:
            instr_tag_name = camel_to_snake(instr.name).upper()
            instr_tag_code.append(f"    {instr_tag_name} = {variant_type}()\n")

        instr_tag_editor.set_with_lock(f"instruction_tag", instr_tag_code)
        module_editor.add_from_import(".instruction_tag", "InstructionTag")

        codegen.package_editor.add_import(
            f"{codegen.root_module}.instructions", "instructions"
        )


class CodeGen:
    """
    Codegen is instantiated for each idl file and generates a client based off that idl.

    To customize client generation, subclass Codegen and override one of the generation functions

    :param idl: The parsed idl file
    :param addresses: Dict from address name to Pubkey for hardcoded associated account addresses
    :param root_module: Name of the root module. If periods are included (e.g. "outer.mid.inner"), multiple directories
                         are created
    :param source_path: Path where the root_module will be output
    :param external_types: Dict from idl typename to function that manages importing the type and returning
                            the in-code typename
    :param default_accounts: Dict from account name to default Pubkey used to populate function
                              argument defaults for convenience
    :param instr_tag_values: Determines how large the instruction tag should be. For anchor programs use "anchor" or omit the arg
    :param accnt_tag_values: Determines how large the account tag should be. For anchor programs use "anchor" or omit the arg
    """
    idl: Idl
    addresses: Dict[str, str]
    root_module: str
    source_path: str
    external_types: Dict[str, Callable]
    default_accounts: Dict[str, Union[Pubkey, str, Callable]]
    instr_tag_values: Literal[
        "anchor",
        "incremental:U8",
        "incremental:U16",
        "incremental:U32",
        "incremental:U64",
        "incremental:U128",
    ]
    accnt_tag_values: Optional[
        Literal[
            "anchor",
            "incremental:U8",
            "incremental:U16",
            "incremental:U32",
            "incremental:U64",
            "incremental:U128",
        ]
    ]
    package_editor: CodeEditor

    _editors: Dict[str, CodeEditor]
    _all_defined_types: Set[str]  # all type names for this module
    _defined_types: Set[str]  # all type names for which code is generated
    _expected_types: Set[str]  # all type names used somewhere in this module

    def __init__(
        self,
        idl,
        addresses,
        root_module,
        source_path,
        external_types: Dict[str, Callable[[CodeEditor], str]] = None,
        default_accounts=None,
        instr_tag_values="incremental[U8]",
        accnt_tag_values="incremental[U8]",
        skip_types=None,
    ):
        self.idl = idl
        self.addresses = addresses
        self.root_module = root_module
        self.source_path = source_path
        self.instr_tag_values = instr_tag_values  # type: ignore
        self.accnt_tag_values = accnt_tag_values  # type: ignore

        if external_types is None:
            self.external_types = dict()
        else:
            self.external_types = external_types

        if default_accounts is None:
            self.default_accounts = dict()
        else:
            self.default_accounts = {
                camel_to_snake(name): val for name, val in default_accounts.items()
            }

        if skip_types is None:
            self.skip_types = []
        else:
            self.skip_types = skip_types

        self._editors = {}
        self._defined_types = set()
        self._expected_types = set()
        self._all_defined_types = set()
        for type_def in self.get_type_definitions():
            self._all_defined_types.add(type_def.name)

    def get_editor(self, name, is_file=True) -> CodeEditor:
        if name not in self._editors:
            subpath = name.replace(".", "/")
            if not is_file:
                subpath += "/__init__"
            subpath += ".py"

            fullpath = os.path.join(self.source_path, subpath)
            self._editors[name] = CodeEditor(fullpath)
            self._editors[name].load()

        return self._editors[name]

    @staticmethod
    def _add_packing_methods(editor, is_struct):
        # this allows IDE's to give better autocomplete for these methods
        # (otherwise, there is no need to add these.)
        if is_struct:
            pass
            # editor.add_lines(
            #     "\n",
            #     "    @classmethod\n",
            #     "    def _to_bytes_partial(cls, buffer, obj):\n",
            #     "        # to modify packing, change this method\n",
            #     "        return super()._to_bytes_partial(buffer, obj)\n",
            #     "\n",
            #     "    @classmethod\n",
            #     "    def _from_bytes_partial(cls, buffer):\n",
            #     "        # to modify unpacking, change this method\n",
            #     "        return super()._from_bytes_partial(buffer)\n",
            # )
        else:
            editor.add_lines(
                "\n",
                "    @classmethod\n",
                "    def _to_bytes_partial(cls, buffer, obj, **kwargs):\n",
                "        # to modify packing, change this method\n",
                "        return super()._to_bytes_partial(buffer, obj, **kwargs)\n",
                "\n",
                "    @classmethod\n",
                "    def _from_bytes_partial(cls, buffer, **kwargs):\n",
                "        # to modify unpacking, change this method\n",
                "        return super()._from_bytes_partial(buffer, **kwargs)\n",
            )

        editor.add_lines(
            "\n" "    @classmethod\n",
            "    def to_bytes(cls, obj, **kwargs):\n",
            '        return cls.pack(obj, converter="bytes", **kwargs)\n',
            "\n",
            "    @classmethod\n",
            "    def from_bytes(cls, raw, **kwargs):\n",
            '        return cls.unpack(raw, converter="bytes", **kwargs)\n',
        )

    def get_type_as_string(
            self, field_type, editor, within_types, explicit_forward_ref=False
    ):
        """
        Convert the idl type to the python type as well as add the required imports
        """
        if field_type.is_a(IdlType.BOOL):
            return "bool"
        elif field_type <= IdlType.I128:
            editor.add_from_import("podite", field_type.get_name())
            return field_type.get_name()
        elif field_type.is_a(IdlType.BYTES):
            return "bytes"
        elif field_type.is_a(IdlType.STRING):
            return "str"
        elif field_type.is_a(IdlType.PUBLIC_KEY):
            editor.add_from_import("solders.pubkey", "Pubkey")
            return "Pubkey"
        elif field_type.is_a(IdlType.DEFINED):
            if field_type.field not in self._all_defined_types:
                if field_type.field in self.external_types:
                    return self.external_types[field_type.field](editor)

            self._expected_types.add(field_type.field)
            if within_types:
                editor.add_from_import(
                    f"{self.root_module}.types.{pascal_to_snake(field_type.field)}",
                    field_type.field,
                )
                if explicit_forward_ref or field_type.field in self._defined_types:
                    return f"{field_type.field}"
                else:
                    return f'"{field_type.field}"'
            else:
                editor.add_from_import(f"{self.root_module}.types", field_type.field)
                return f"{field_type.field}"
        elif int(field_type) in (
            int(IdlType.OPTION),
            int(IdlType.COPTION),
            int(IdlType.STATIC),
            int(IdlType.VEC),
        ):
            if field_type.is_a(IdlType.COPTION):
                type_name = "COption"
                editor.add_from_import(f"dexteritysdk.solmate.dtypes", type_name)
            else:
                if field_type.is_a(IdlType.OPTION):
                    type_name = "Option"
                elif field_type.is_a(IdlType.STATIC):
                    type_name = "Static"
                else:
                    type_name = "Vec"

                editor.add_from_import(f"podite", type_name)

            field_type_str = self.get_type_as_string(
                field_type.field, editor, within_types=within_types
            )
            return f"{type_name}[{field_type_str}]"

        elif field_type.is_a(IdlType.ARRAY):
            editor.add_from_import("podite", "FixedLenArray")
            elem_type, n_elem = field_type.field
            elem_type_str = self.get_type_as_string(
                elem_type, editor, within_types=within_types
            )
            return f"FixedLenArray[{elem_type_str}, {n_elem}]"

        return ""

    def get_type_definitions(self) -> Iterable[IdlTypeDefinition]:
        return self.idl.types + self.idl.accounts

    def generate_types(self):
        """
        Method that generates python files for each type defined in the idl
        """
        type_definitions = self.get_type_definitions()
        if not type_definitions:
            return

        module_editor = self.get_editor(f"{self.root_module}.types", is_file=False)
        for type_def in type_definitions:
            editor = self.get_editor(
                f"{self.root_module}.types.{camel_to_snake(type_def.name)}"
            )
            editor.add_from_import("podite", "pod")

            class_code = ["@pod\n"]
            if type_def.type.is_a(IdlTypeDefinitionTy.STRUCT):
                class_code += [f"class {type_def.name}:\n"]
                has_field = False
                for field in type_def.type.field.fields:
                    field_name = camel_to_snake(field.name)
                    field_type = self.get_type_as_string(
                        field.type, editor, within_types=True
                    )
                    class_code.append(f"    {field_name}: {field_type}\n")
                    has_field = True

                if not has_field:
                    class_code += ["    pass\n"]

            else:
                editor.add_from_import("podite", "Enum")
                editor.add_from_import("podite", "AutoTagType")
                class_code += [f"class {type_def.name}(Enum[AutoTagType]):\n"]
                variants = type_def.type.field.variants
                for variant in variants:
                    if variant.fields is None:
                        variant_type = "None"
                    elif variant.fields.is_a(EnumFields.NAMED):
                        editor.add_from_import("podite", "Variant")
                        editor.add_from_import("podite", "named_fields")
                        editor.add_from_import("podite", "Option")

                        named_fields = []
                        for field in variant.fields.field:
                            field_name = field.name
                            field_type = self.get_type_as_string(
                                field.type,
                                editor,
                                within_types=True,
                                explicit_forward_ref=True,
                            )
                            named_fields.append(f"{field_name}={field_type}")

                        if len(named_fields) == 0:
                            variant_type = "None"
                        else:
                            variant_type = "Variant(field=named_fields(" + ", ".join(named_fields) + "))"
                    else:
                        editor.add_from_import("podite", "Variant")

                        tuple_fields = []
                        for field_type in variant.fields.field:
                            tuple_fields.append(
                                self.get_type_as_string(
                                    field_type,
                                    editor,
                                    within_types=True,
                                    explicit_forward_ref=True,
                                )
                            )

                        if len(tuple_fields) == 0:
                            variant_type = "None"
                        elif len(tuple_fields) == 1:
                            variant_type = "Variant(field=" + tuple_fields[0] + ")"
                        else:
                            editor.add_from_import("typing", "Tuple")
                            variant_type = (
                                "Variant(field=Tuple[" + ", ".join(tuple_fields) + "])"
                            )

                    variant_name = pascal_to_snake(variant.name).upper()
                    class_code += [f"    {variant_name} = {variant_type}\n"]

                if not variants:
                    class_code += ["    pass\n"]

            if editor.set_with_lock(f"class({type_def.name})", class_code):
                self._add_packing_methods(
                    editor, type_def.type.is_a(IdlTypeDefinitionTy.STRUCT)
                )

            module_editor.add_from_import(
                f".{camel_to_snake(type_def.name)}", type_def.name
            )

            self._defined_types.add(type_def.name)
            self.package_editor.add_import(f"{self.root_module}.types", "types")

    def generate_addresses(self):
        editor = self.get_editor(f"{self.root_module}.addrs")

        editor.add_from_import("solders.pubkey", "Pubkey")

        code = []
        for name, addr in self.addresses.items():
            code.append(f'{name} = Pubkey.from_string("{addr}")\n')
        editor.set_with_lock("addresses", code)

        program_id = self.get_account_address(self.package_editor, "program_id")
        if program_id is None:
            raise RuntimeError("Default accounts must contain program_id.")

    def generate_constants(self):
        if not self.idl.constants:
            return

        editor = self.get_editor(f"{self.root_module}.constants")
        code = []

        for const in self.idl.constants:
            const_type = self.get_type_as_string(
                const.type, editor, within_types=False
            )
            code.append(f"{const.name}: {const_type} = {const.value}\n")

        editor.set_with_lock("constants", code)
        self.package_editor.add_import(f"{self.root_module}.constants", "constants")

    def generate_accounts(self):
        if not self.idl.accounts or self.accnt_tag_values is None:
            return

        editor = self.get_editor(f"{self.root_module}.accounts")
        code = []

        editor.add_from_import("podite", "pod")
        editor.add_from_import("podite", "Enum")

        base = None
        variant_type = None

        if self.accnt_tag_values == "anchor":
            editor.add_from_import("podite", "U64")
            editor.add_from_import("dexteritysdk.solmate.anchor", "AccountDiscriminant")
            base = "Enum[U64]"
            variant_type = "AccountDiscriminant"

        elif self.accnt_tag_values is not None:
            tag_type = self.instr_tag_values.split(":")[1]
            editor.add_from_import("podite", "Variant")
            editor.add_from_import("podite", tag_type)
            base = f"Enum[{tag_type}]"
            variant_type = "Variant"

        code.extend(
            [
                "@pod\n",
                f"class Accounts({base}):\n",
            ]
        )
        for account in self.idl.accounts:
            editor.add_from_import(
                f"{self.root_module}.types.{pascal_to_snake(account.name)}",
                account.name,
            )

            variant_name = pascal_to_snake(account.name).upper()
            variant_inst = f"{variant_type}(field={account.name})"
            code += [f"    {variant_name} = {variant_inst}\n"]

        if editor.set_with_lock("accounts", code):
            self._add_packing_methods(editor, is_struct=True)

        self.package_editor.add_import(f"{self.root_module}.accounts", "accounts")

    def get_account_address(self, editor, account_name):
        default_account = self.default_accounts.get(account_name, None)
        if default_account is not None:
            if isinstance(default_account, (str, Pubkey)):
                expr = f'Pubkey.from_string("{default_account}")'
            else:
                expr = default_account(editor)
            return expr
        return None

    def generate_instructions(self):
        return InstructionCodeGen(self).generate()  # here

    def generate_events(self):
        if not self.idl.events:
            return

        # TODO implement code generation for events
        print("Skipping events...")

    def generate_errors(self):
        if not self.idl.errors:
            return

        code = ["@pod\n", "class Error(Enum[U64]):\n"]
        for error in self.idl.errors:
            error_name = pascal_to_snake(error.name).upper()
            code.append(
                f'    {error_name} = Variant({error.code}, field="{error.msg}")\n'
            )

        editor = self.get_editor(f"{self.root_module}.errors")
        editor.add_from_import("podite", "pod")
        editor.add_from_import("podite", "Enum")
        editor.add_from_import("podite", "Variant")
        editor.add_from_import("podite", "U64")
        if editor.set_with_lock("errors", code):
            editor.add_lines(
                "\n",
                "    @classmethod\n",
                "    def to_bytes(cls, obj, **kwargs):\n",
                '        return cls.pack(obj, converter="bytes", **kwargs)\n',
                "\n",
                "    @classmethod\n",
                "    def from_bytes(cls, raw, **kwargs):\n",
                '        return cls.unpack(raw, converter="bytes", **kwargs)\n',
            )

        self.package_editor.add_from_import(f"{self.root_module}.errors", "Error")

    def generate_state(self):
        if not self.idl.state:
            return

        # TODO implement code generation for state
        print("Skipping state...")

    def generate_code(self, check_missing_types=False):
        self.package_editor = self.get_editor(self.root_module, is_file=False)

        self.generate_addresses()
        self.generate_types()
        self.generate_constants()
        self.generate_accounts()
        self.generate_instructions()
        self.generate_events()
        self.generate_errors()
        self.generate_state()

        undefined_types = self._expected_types.difference(self._defined_types)
        if undefined_types:
            if check_missing_types:
                raise RuntimeError(
                    f"The following types are used but not defined: {undefined_types}"
                )
            else:
                print(
                    f"The following types are used but not defined: {undefined_types}"
                )

    def save_modules(self):
        for editor in self._editors.values():
            editor.save()


#########################
# Common external types #
#########################

def usize_type(editor: CodeEditor):
    editor.add_from_import("dexteritysdk.solmate.dtypes", "Usize")
    return "Usize"


def unix_timestamp_type(editor: CodeEditor):
    editor.add_from_import("dexteritysdk.solmate.dtypes", "UnixTimestamp")
    return "UnixTimestamp"


def program_error_type(editor: CodeEditor):
    editor.add_from_import("dexteritysdk.solmate.dtypes", "ProgramError")
    return "ProgramError"


def cli(
    idl_path: str,
    addresses: Dict[str, str],
    source_path: str,
    root_module: str,
    skip_types: Set[str],
    default_accounts: Dict[str, str],
    instr_tag_values: Literal["anchor", "incremental"],
    accnt_tag_values: Literal["anchor", "incremental"],
    external_types: Dict[str, Callable[[CodeEditor], str]] = None,
):
    idl = Idl.from_json_file(idl_path)
    idl.types = list(filter(lambda x: x.name not in skip_types, idl.types))

    external_types = {
        "usize": usize_type,
        "UnixTimestamp": unix_timestamp_type,
        "ProgramError": program_error_type,
    }.update(external_types if external_types is not None else {})

    print(f"Generating code for {idl_path}...")
    codegen = CodeGen(
        idl=idl,
        addresses=addresses,
        root_module=root_module,
        source_path=source_path,
        external_types=external_types,
        default_accounts=default_accounts,
        instr_tag_values=instr_tag_values,
        accnt_tag_values=accnt_tag_values,
        skip_types=skip_types,
    )
    codegen.generate_code(check_missing_types=not True)
    codegen.save_modules()



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.risk_engine.types.fast_int import FastInt
from dexteritysdk.codegen.risk_engine.types.risk_account_tag import RiskAccountTag
from dexteritysdk.solmate.dtypes import Usize
from podite import (
    FixedLenArray,
    U64,
    pod,
)
from solders.pubkey import Pubkey

# LOCK-END


# LOCK-BEGIN[class(CovarianceMetadata)]: DON'T MODIFY
@pod
class CovarianceMetadata:
    tag: RiskAccountTag
    update_slot: U64
    authority: Pubkey
    num_active_products: Usize
    product_keys: FixedLenArray[Pubkey, 128]
    standard_deviations: FixedLenArray[FastInt, 128]
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .create_risk_state_account import (
    CreateRiskStateAccountIx,
    create_risk_state_account,
)
from .instruction_tag import InstructionTag
from .validate_account_health import (
    ValidateAccountHealthIx,
    validate_account_health,
)
from .validate_account_liquidation import (
    ValidateAccountLiquidationIx,
    validate_account_liquidation,
)

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    FixedLenArray,
    I16,
    U64,
    U8,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(InitializeMarketProductGroupParams)]: DON'T MODIFY
@pod
class InitializeMarketProductGroupParams:
    name: FixedLenArray[U8, 16]
    validate_account_discriminant_len: U64
    find_fees_discriminant_len: U64
    validate_account_health_discriminant: FixedLenArray[U8, 8]
    validate_account_liquidation_discriminant: FixedLenArray[U8, 8]
    create_risk_state_account_discriminant: FixedLenArray[U8, 8]
    create_fee_state_account_discriminant: FixedLenArray[U8, 8]
    close_risk_state_account_discriminant: FixedLenArray[U8, 8]
    close_fee_state_account_discriminant: FixedLenArray[U8, 8]
    find_fees_discriminant: FixedLenArray[U8, 8]
    max_maker_fee_bps: I16
    min_maker_fee_bps: I16
    max_taker_fee_bps: I16
    min_taker_fee_bps: I16
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dexteritysdk',
 'dexteritysdk.bots',
 'dexteritysdk.codegen',
 'dexteritysdk.codegen.alpha_risk_engine',
 'dexteritysdk.codegen.alpha_risk_engine.instructions',
 'dexteritysdk.codegen.alpha_risk_engine.types',
 'dexteritysdk.codegen.dex',
 'dexteritysdk.codegen.dex.instructions',
 'dexteritysdk.codegen.dex.types',
 'dexteritysdk.codegen.instruments',
 'dexteritysdk.codegen.instruments.instructions',
 'dexteritysdk.codegen.instruments.types',
 'dexteritysdk.codegen.noop_risk_engine',
 'dexteritysdk.codegen.noop_risk_engine.instructions',
 'dexteritysdk.codegen.risk_engine',
 'dexteritysdk.codegen.risk_engine.instructions',
 'dexteritysdk.codegen.risk_engine.types',
 'dexteritysdk.constant_fees',
 'dexteritysdk.dex',
 'dexteritysdk.dummy_oracle',
 'dexteritysdk.dummy_oracle.instructions',
 'dexteritysdk.dummy_oracle.state',
 'dexteritysdk.instruments',
 'dexteritysdk.instruments.instructions',
 'dexteritysdk.pyserum',
 'dexteritysdk.pyserum.layouts',
 'dexteritysdk.pyserum.structs',
 'dexteritysdk.risk',
 'dexteritysdk.scripts',
 'dexteritysdk.solmate',
 'dexteritysdk.solmate.anchor',
 'dexteritysdk.utils',
 'dexteritysdk.utils.aob',
 'dexteritysdk.utils.aob.state']

package_data = \
{'': ['*'], 'dexteritysdk': ['idl/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'MarkupSafe>=2.0.1,<3.0.0',
 'OSlash>=0.6.3,<0.7.0',
 'Pillow>=9.0.1,<10.0.0',
 'PyNaCl>=1.5.0,<2.0.0',
 'Pygments>=2.11.2,<3.0.0',
 'QtPy>=2.0.1,<3.0.0',
 'Send2Trash>=1.8.0,<2.0.0',
 'aiodns>=3.0.0,<4.0.0',
 'aiohttp>=3.8.1,<4.0.0',
 'aiosignal>=1.2.0,<2.0.0',
 'anchorpy>=0.18.0,<0.19.0',
 'anyio>=3.5.0,<4.0.0',
 'appnope>=0.1.2,<0.2.0',
 'argon2-cffi>=21.3.0,<22.0.0',
 'argparse>=1.4.0,<2.0.0',
 'async-timeout>=4.0.2,<5.0.0',
 'attrs>=21.4.0,<22.0.0',
 'backcall>=0.2.0,<0.3.0',
 'backoff>=1.11.1,<2.0.0',
 'base58>=2.1.1,<3.0.0',
 'bleach>=4.1.0,<5.0.0',
 'borsh-construct>=0.1.0,<0.2.0',
 'certifi>=2021.10.8,<2022.0.0',
 'cffi>=1.15.0,<2.0.0',
 'charset-normalizer>=2.0.12,<3.0.0',
 'construct-typing>=0.5.2,<0.6.0',
 'construct==2.10.68',
 'cryptography>=36.0.1,<37.0.0',
 'cycler>=0.11.0,<0.12.0',
 'debugpy>=1.5.1,<2.0.0',
 'decorator>=5.1.1,<6.0.0',
 'defusedxml>=0.7.1,<0.8.0',
 'dnspython>=2.2.0,<3.0.0',
 'entrypoints>=0.4,<0.5',
 'flake8>=4.0.1,<5.0.0',
 'frozenlist>=1.3.0,<2.0.0',
 'h11==0.12.0',
 'idna>=3.3,<4.0',
 'ipykernel>=6.9.0,<7.0.0',
 'ipython>=8.0.1,<9.0.0',
 'ipython_genutils>=0.2.0,<0.3.0',
 'ipywidgets>=7.6.5,<8.0.0',
 'jedi>=0.18.1,<0.19.0',
 'jsonrpcclient>=4.0.2,<5.0.0',
 'jsonrpcserver>=5.0.8,<6.0.0',
 'jsonschema==3.2.0',
 'jupyter-client>=7.1.2,<8.0.0',
 'jupyter-console>=6.4.0,<7.0.0',
 'jupyter-core>=4.9.1,<5.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'jupyterlab-pygments>=0.1.2,<0.2.0',
 'jupyterlab-widgets>=1.0.2,<2.0.0',
 'kiwisolver>=1.3.2,<2.0.0',
 'loguru>=0.6.0,<0.7.0',
 'lz4==4.3.3',
 'matplotlib-inline>=0.1.3,<0.2.0',
 'matplotlib>=3.5.1,<4.0.0',
 'mccabe==0.6.1',
 'mistune==0.8.4',
 'multidict>=6.0.2,<7.0.0',
 'nbclient>=0.5.10,<0.6.0',
 'nbconvert==6.2',
 'nbformat>=5.1.3,<6.0.0',
 'nest-asyncio>=1.5.4,<2.0.0',
 'notebook>=6.4.8,<7.0.0',
 'numpy>=1.22.2,<2.0.0',
 'packaging>=21.3,<22.0',
 'pandas>=1.4.1,<2.0.0',
 'pandocfilters>=1.5.0,<2.0.0',
 'parso>=0.8.3,<0.9.0',
 'pexpect>=4.8.0,<5.0.0',
 'pickleshare>=0.7.5,<0.8.0',
 'podite==0.1.2',
 'prometheus-client>=0.13.1,<0.14.0',
 'prompt-toolkit>=3.0.28,<4.0.0',
 'ptyprocess>=0.7.0,<0.8.0',
 'pycares>=4.1.2,<5.0.0',
 'pycodestyle>=2.8.0,<3.0.0',
 'pycparser>=2.21,<3.0',
 'pyflakes>=2.4.0,<3.0.0',
 'pyparsing>=3.0.7,<4.0.0',
 'pyrsistent>=0.18.1,<0.19.0',
 'pythclient>=0.1.2,<0.2.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'pytz>=2022.1,<2023.0',
 'pyzmq>=22.3.0,<23.0.0',
 'qtconsole>=5.2.2,<6.0.0',
 'requests>=2.27.1,<3.0.0',
 'rfc3986==1.5',
 'six>=1.16.0,<2.0.0',
 'sniffio>=1.2.0,<2.0.0',
 'solana>=0.30.2,<0.31.0',
 'terminado>=0.13.1,<0.14.0',
 'testpath>=0.5.0,<0.6.0',
 'tornado>=6.1,<7.0',
 'traitlets>=5.1.1,<6.0.0',
 'types-cachetools>=4.2.9,<5.0.0',
 'typing-extensions==4.2.0',
 'urllib3>=1.26.8,<2.0.0',
 'wcwidth>=0.2.5,<0.3.0',
 'webencodings>=0.5.1,<0.6.0',
 'websockets>=10.1,<11.0',
 'widgetsnbextension>=3.5.2,<4.0.0',
 'yarl>=1.7.2,<2.0.0']

entry_points = \
{'console_scripts': ['bootstrap = dexteritysdk.bots.bootstrap:main',
                     'connect = dexteritysdk.bots.new_connect:main',
                     'discriminant = dexteritysdk.scripts.discriminant:main',
                     'extract-program-ids = '
                     'dexteritysdk.scripts.extract_program_ids:main',
                     'generate-code = dexteritysdk.scripts.generate_code:main',
                     'get-trader-risk-groups = '
                     'dexteritysdk.scripts.get_trader_risk_groups:main',
                     'keypair-to-b58 = '
                     'dexteritysdk.scripts.keypair_to_b58:main',
                     'local-validator = '
                     'dexteritysdk.scripts.start_test_validator:main']}

setup_kwargs = {
    'name': 'dexteritysdk',
    'version': '0.2.50',
    'description': 'Client for Dexterity - a modular derivatives decentralized exchange reference implementation',
    'long_description': '# Dexterity client SDK\n**Dexterity** is a derivatives decentralized exchange running on Solana.\nThis package provides the basic blocks in order to integrate and trade on Dexterity.\n\nYou can learn more about Dexterity [here](https://docs.hxro.network/market-protocols/derivatives-protocol/dexterity).\n\n## Prerequisites\n* [Solana.py](https://pypi.org/project/solana/)\n\n## Creating the client\nFirst, an instance of a Solana client is required.\n```python\nfrom dexteritysdk.dex.sdk_context import SDKContext, SDKTrader\nfrom solana.rpc.api import Client\nfrom solana.keypair import Keypair\nfrom solana.pubkey import Pubkey\n\ndef main():\n\t# use "https://api.mainnet-beta.solana.com" for mainnet\n\tnetwork = "https://api.devnet.solana.com"\n\tclient = Client(network)\n```\n\nA Keypair for the payer has to also be provided.\n```python\nkeypair = Keypair.from_secret_key(keypair_bytes)\n```\nThe SDK also requires the **Market Product Group (MPG)** that we\'re going to trade on.\n\nThe public key of the default MPG is **`HiCy6vzuN3yLXD3z35D6nV7bzNLcyrvGLf3uSKuutSLo`** for mainnet or **`HyWxreWnng9ZBDPYpuYugAfpCMkRkJ1oz93oyoybDFLB`** for devnet.\n```python\n# mainnet MPG\nmpg = Pubkey("HiCy6vzuN3yLXD3z35D6nV7bzNLcyrvGLf3uSKuutSLo")\n\n# ** OR **\n# devnet MPG\nmpg = Pubkey("HyWxreWnng9ZBDPYpuYugAfpCMkRkJ1oz93oyoybDFLB")\n```\nNow, we\'re ready to get an instance of `SDKContext` for this MPG.\n```python\nctx = SDKContext.connect(\n\tclient=client,\n\tmarket_product_group_key=mpg,\n\tpayer=keypair,\n\traise_on_error=True\n)\n```\n\n## Registering new trader / TRG\nIf you\'re a new trader, creation of a trading account is required.\n```python\ntrader = ctx.register_trader(keypair)\ntrg_key = trader.account\n```\n\nThe **Trader Risk Group (TRG)** is your trading account, that you can now use.\n\n## Using an existing TRG\nIf you already have a trader set up, you need to provide the TRG public key for it.\nThe SDK allows you to list all your registered TRGs.\n```python\n# this will return a list of your all your TRG public keys\n# for example: [HeykeQWRh6DC2Tz5X3WBuWdMHicyECDEFGMjomV6LBye, HeyZNJ9gQVAEqHeCFFQ781E53d66DATKXHeynwnCFBye]\ntrg_keys = ctx.list_trader_risk_groups()\n```\n\nAfter finding your TRG, you can initialise the `SDKTrader` instance.\n```python\nfrom dexteritysdk.codegen.dex.types import TraderRiskGroup\nfrom dexteritysdk.program_ids import *\nfrom dexteritysdk.utils.solana import explore\n\n# ...\ntrader = SDKTrader.connect(ctx, trg_pubkey, keypair)\n```\n\n## Funding the TRG\nTo start trading, you need to deposit funds.\nThe tokens deposited have to be MPG vault mint tokens (e.g. USDC).\nThe wallet you provided will be debited.\n```python\n# this will deposit 733.1 USDC\ntrader.deposit(ctx, 733.1)\n```\n\nYou can also withdraw in a similar fashion.\n```python\n# this will withdraw 0.1 USDC back to your wallet\ntrader.withdraw(ctx, 0.1)\n```\n\nYou can check your balance at any time.\n```python\ntrg, _ = trader.get_trader_risk_group()\ntrg.cash_balance\n```\n\n## Trading on Dexterity\n\n### Listing trading products and order books\nThe `products` field of a `SDKContext` instance, is a list of the available trading products (`SDKProduct`).\nEvery `SDKProduct` has a `name` and a `get_orderbook(ctx)` function that will return its order book (`SDKOrderBook`).\nThe `SDKOrderBook` contains all `SDKOrders` on each side of the book.\n\nHere\'s an example:\n```python\nfor product in ctx.products:\n\tbook = product.get_orderbook(ctx, refresh=True)\n\tprint(f"Printing the order book of {product.name}")\n\tfor order in book.bids:\n\t\tprint(f"Bid of size {order.qty} at {order.price}")\n\tfor order in book.asks:\n\t\tprint(f"Ask of size {order.qty} at {order.price}")\n```\n\n### Placing an order\n\nPlacing an order is straightforward.\n```python\nfrom dexteritysdk.utils.aob import Side\n# ...\norder_summary = trader.place_order(ctx, product, Side.ASK, size, price)\n```\nThe returned `SDKOrderSummary` contains the `order_id` of the new order, as well as the `remaining_qty` and `filled_qty` if your order has been filled. If the order is immediately fully filled, `order_id` will be `None`.\n\nIf there\'s any error, an exception will be raised.\n\n### Cancelling orders\nTo cancel a previous order you can call `trader.cancel(ctx, product, order_id)`.\nFollowing the previous example:\n```python\ntrader.cancel(ctx, product, order_summary.order_id)\n```\n\nYou can also cancel all your open orders for multiple products.\n```python\n# products is a list of SDKProduct\n# if products list is empty or None, then all your open orders will be cancelled\nproducts = [ctx.products[0], ctx.products[2]]\ntrader.cancel_all_orders(ctx, products)\n```\nIf any cancel fails, an exception will be raised.\n\n### Positions\nPositions can also be listed, by iterating the `trader_positions` array:\n```python\ntrg, _ = trader.get_trader_risk_group()\ntrg.trader_positions\n```\n',
    'author': 'Joe Howarth',
    'author_email': 'josephehowarth@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solders.instruction import (
    AccountMeta,
    Instruction,
)
from solders.pubkey import Pubkey
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(settle_derivative)]: DON'T MODIFY
@dataclass
class SettleDerivativeIx:
    program_id: Pubkey

    # account metas
    market_product_group: AccountMeta
    derivative_metadata: AccountMeta
    price_oracle: AccountMeta
    dex_program: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    def to_instruction(self):
        keys = []
        keys.append(self.market_product_group)
        keys.append(self.derivative_metadata)
        keys.append(self.price_oracle)
        keys.append(self.dex_program)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.SETTLE_DERIVATIVE))

        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(settle_derivative)]: DON'T MODIFY
def settle_derivative(
    market_product_group: Union[str, Pubkey, AccountMeta],
    derivative_metadata: Union[str, Pubkey, AccountMeta],
    price_oracle: Union[str, Pubkey, AccountMeta],
    dex_program: Union[str, Pubkey, AccountMeta],
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[Pubkey] = None,
):
    if program_id is None:
        program_id = Pubkey.from_string("8981bZYszfz1FrFVx7gcUm61RfawMoAHnURuERRJKdkq")

    if isinstance(market_product_group, (str, Pubkey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(derivative_metadata, (str, Pubkey)):
        derivative_metadata = to_account_meta(
            derivative_metadata,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(price_oracle, (str, Pubkey)):
        price_oracle = to_account_meta(
            price_oracle,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(dex_program, (str, Pubkey)):
        dex_program = to_account_meta(
            dex_program,
            is_signer=False,
            is_writable=False,
        )

    return SettleDerivativeIx(
        program_id=program_id,
        market_product_group=market_product_group,
        derivative_metadata=derivative_metadata,
        price_oracle=price_oracle,
        dex_program=dex_program,
        remaining_accounts=remaining_accounts,
    ).to_instruction()

# LOCK-END



"""Slab data stucture that is used to represent Order book."""
from __future__ import annotations

from enum import IntEnum

from construct import Bytes, Int8ul, Int32ul, Int64ul, Padding
from construct import Struct as cStruct
from construct import Switch
from borsh_construct import CStruct, I8, I32, I64

KEY = Bytes(16)

SLAB_HEADER_LAYOUT = CStruct(
    "tag" / I8,
    "bump_index" / I64,
    "free_list_length" / I64,
    "free_list_head" / I32,
    "callback_memory_offset" / I64,
    "callback_free_list_len" / I64,
    "callback_free_list_head" / I64,
    "callback_bump_index" / I64,
    "root" / I32,
    "leaf_count" / I64,
    "market" / Bytes(32),
)


class SlabType(IntEnum):
    UNINITIALIZED = 0
    MARKET = 1
    EVENT_QUEUE = 2
    BIDS = 3
    ASKS = 4


class NodeType(IntEnum):
    UNINITIALIZED = 0
    INNER_NODE = 1
    LEAF_NODE = 2
    FREE_NODE = 3
    LAST_FREE_NODE = 4


# Different node types, we pad it all to size of 32 bytes.
UNINITIALIZED = cStruct(Padding(32))
INNER_NODE = cStruct("prefix_len" / Int64ul, "key" / KEY, "children" / Int32ul[2])
LEAF_NODE = cStruct(
    "key" / KEY,
    "callback" / Int32ul,
    "expiration_slot" / Int32ul,
    "quantity" / Int64ul,
)
FREE_NODE = cStruct("next" / Int32ul, Padding(28))
LAST_FREE_NODE = cStruct(Padding(32))

SLAB_NODE_LAYOUT = cStruct(
    "tag" / Int8ul,
    Padding(7),
    "node"
    / Switch(
        lambda this: this.tag,
        {
            NodeType.UNINITIALIZED: UNINITIALIZED,
            NodeType.INNER_NODE: INNER_NODE,
            NodeType.LEAF_NODE: LEAF_NODE,
            NodeType.FREE_NODE: FREE_NODE,
            NodeType.LAST_FREE_NODE: LAST_FREE_NODE,
        },
    ),
)

SLAB_LAYOUT = cStruct("header" / SLAB_HEADER_LAYOUT, Padding(7),
                      "nodes" / SLAB_NODE_LAYOUT[lambda this: this.header.bump_index])



__version__ = "0.1.0"

from typing import Any

from podite import get_catalog
from podite.bytes import BytesPodConverter
from podite.json import JsonPodConverter
from solders.pubkey import Pubkey


def register_solana_types_converters():
    class SolanaTypesPodConverter(BytesPodConverter, JsonPodConverter):
        def get_mapping(self, type_):
            if type_ in (Pubkey,):
                return self

            return None

        def is_static(self, type_) -> bool:
            if type_ == Pubkey:
                return True

            raise ValueError()

        def calc_size(self, type_, obj, **kwargs) -> int:
            if type_ == Pubkey:
                return 32

            raise ValueError()

        def calc_max_size(self, type_) -> int:
            if type_ == Pubkey:
                return 32

            raise ValueError()

        def pack_partial(self, type_, buffer, obj, **kwargs) -> Any:
            if type_ == Pubkey:
                return buffer.write(bytes(obj))

            raise ValueError()

        def unpack_partial(self, type_, buffer, **kwargs) -> Any:
            if type_ == Pubkey:
                return Pubkey(buffer.read(32))

            raise ValueError()

        def pack_dict(self, type_, obj, **kwargs) -> Any:
            if type_ == Pubkey:
                return str(obj)

            raise ValueError()

        def unpack_dict(self, type_, obj, **kwargs) -> Any:
            if type_ == Pubkey:
                if not isinstance(obj, str):
                    raise ValueError("Pubkey has to be encoded as a string.")

                return Pubkey.from_string(obj)

            raise ValueError()

    converter = SolanaTypesPodConverter()
    get_catalog("bytes").register(converter.get_mapping)
    get_catalog("json").register(converter.get_mapping)


register_solana_types_converters()



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.risk_engine.types.risk_account_tag import RiskAccountTag
from dexteritysdk.solmate.dtypes import Usize
from podite import (
    FixedLenArray,
    I8,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(CorrelationMatrix)]: DON'T MODIFY
@pod
class CorrelationMatrix:
    tag: RiskAccountTag
    num_active_products: Usize
    correlations: FixedLenArray[I8, 8256]
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from solders.pubkey import Pubkey

from dexteritysdk.dummy_oracle.state.common import AccountTag
from podite import pod, U64, I64


@pod
class OraclePrice:
    tag: AccountTag
    price: I64
    decimals: U64
    slot: U64
    update_authority: Pubkey



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solders.instruction import (
    AccountMeta,
    Instruction,
)
from solders.pubkey import Pubkey
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(initialize_trader_risk_group)]: DON'T MODIFY
@dataclass
class InitializeTraderRiskGroupIx:
    program_id: Pubkey

    # account metas
    owner: AccountMeta
    trader_risk_group: AccountMeta
    market_product_group: AccountMeta
    risk_signer: AccountMeta
    trader_risk_state_acct: AccountMeta
    trader_fee_state_acct: AccountMeta
    risk_engine_program: AccountMeta
    fee_model_config_acct: AccountMeta
    fee_model_program: AccountMeta
    system_program: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    def to_instruction(self):
        keys = []
        keys.append(self.owner)
        keys.append(self.trader_risk_group)
        keys.append(self.market_product_group)
        keys.append(self.risk_signer)
        keys.append(self.trader_risk_state_acct)
        keys.append(self.trader_fee_state_acct)
        keys.append(self.risk_engine_program)
        keys.append(self.fee_model_config_acct)
        keys.append(self.fee_model_program)
        keys.append(self.system_program)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.INITIALIZE_TRADER_RISK_GROUP))

        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(initialize_trader_risk_group)]: DON'T MODIFY
def initialize_trader_risk_group(
    owner: Union[str, Pubkey, AccountMeta],
    trader_risk_group: Union[str, Pubkey, AccountMeta],
    market_product_group: Union[str, Pubkey, AccountMeta],
    risk_signer: Union[str, Pubkey, AccountMeta],
    trader_risk_state_acct: Union[str, Pubkey, AccountMeta],
    trader_fee_state_acct: Union[str, Pubkey, AccountMeta],
    risk_engine_program: Union[str, Pubkey, AccountMeta],
    fee_model_config_acct: Union[str, Pubkey, AccountMeta],
    fee_model_program: Union[str, Pubkey, AccountMeta],
    system_program: Union[str, Pubkey, AccountMeta] = Pubkey.from_string("11111111111111111111111111111111"),
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[Pubkey] = None,
):
    if program_id is None:
        program_id = Pubkey.from_string("FUfpR31LmcP1VSbz5zDaM7nxnH55iBHkpwusgrnhaFjL")

    if isinstance(owner, (str, Pubkey)):
        owner = to_account_meta(
            owner,
            is_signer=True,
            is_writable=True,
        )
    if isinstance(trader_risk_group, (str, Pubkey)):
        trader_risk_group = to_account_meta(
            trader_risk_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(market_product_group, (str, Pubkey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_signer, (str, Pubkey)):
        risk_signer = to_account_meta(
            risk_signer,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(trader_risk_state_acct, (str, Pubkey)):
        trader_risk_state_acct = to_account_meta(
            trader_risk_state_acct,
            is_signer=True,
            is_writable=True,
        )
    if isinstance(trader_fee_state_acct, (str, Pubkey)):
        trader_fee_state_acct = to_account_meta(
            trader_fee_state_acct,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(risk_engine_program, (str, Pubkey)):
        risk_engine_program = to_account_meta(
            risk_engine_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_model_config_acct, (str, Pubkey)):
        fee_model_config_acct = to_account_meta(
            fee_model_config_acct,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_model_program, (str, Pubkey)):
        fee_model_program = to_account_meta(
            fee_model_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(system_program, (str, Pubkey)):
        system_program = to_account_meta(
            system_program,
            is_signer=False,
            is_writable=False,
        )

    return InitializeTraderRiskGroupIx(
        program_id=program_id,
        owner=owner,
        trader_risk_group=trader_risk_group,
        market_product_group=market_product_group,
        risk_signer=risk_signer,
        trader_risk_state_acct=trader_risk_state_acct,
        trader_fee_state_acct=trader_fee_state_acct,
        risk_engine_program=risk_engine_program,
        fee_model_config_acct=fee_model_config_acct,
        fee_model_program=fee_model_program,
        system_program=system_program,
        remaining_accounts=remaining_accounts,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    FixedLenArray,
    U64,
    pod,
)
from solders.pubkey import Pubkey
from dexteritysdk.solmate.dtypes import UnixTimestamp

# LOCK-END


# LOCK-BEGIN[class(InitializeFixedIncomeParams)]: DON'T MODIFY
@pod
class InitializeFixedIncomeParams:
    face_value: U64
    coupon_rate: U64
    initialization_time: UnixTimestamp
    coupon_dates: FixedLenArray[UnixTimestamp, 32]
    maturity_date: UnixTimestamp
    close_authority: Pubkey
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)
